from abc import ABC, abstractmethod

from rastervision.data import ActivateMixin


class LabelStore(ABC, ActivateMixin):
    """This defines how to store prediction labels are stored for a scene.
    """

    @abstractmethod
    def save(self, labels):
        """Save.

        Args:
           labels - Labels to be saved, the type of which will be dependant on the type
                    of task.
        """
        pass

    @abstractmethod
    def get_labels(self):
        """Loads Labels from this label store."""
        pass

    @abstractmethod
    def empty_labels(self):
        """Produces an empty Labels"""
        pass

    def _subcomponents_to_activate(self):
        pass

    def _activate(self):
        pass

    def _deactivate(self):
        pass



from abc import (ABC, abstractmethod)

import rastervision as rv


class RasterSourceDefaultProvider(ABC):
    @staticmethod
    @abstractmethod
    def handles(s):
        """Returns True if this provider is a default for this string"""
        pass

    @abstractmethod
    def construct(s, channel_order=None):
        """Constructs default based on the string and an optional channel order."""
        pass


class RasterioSourceDefaultProvider(RasterSourceDefaultProvider):
    @staticmethod
    def handles(uri):
        # Since there are so many types handled by Rasterio/GDAL, the RasterioSource
        # will be the catch-all. More specific types can be handled by other
        # RasterSources.
        return True

    @staticmethod
    def construct(uri, channel_order=None):
        return rv.RasterSourceConfig.builder(rv.RASTERIO_SOURCE) \
                                    .with_uri(uri) \
                                    .with_channel_order(channel_order) \
                                    .build()



from rastervision.analyzer import Analyzer
from rastervision.core import RasterStats


class StatsAnalyzer(Analyzer):
    """Computes RasterStats against the entire scene set.
    """

    def __init__(self, stats_uri, sample_prob=None):
        self.stats_uri = stats_uri
        self.sample_prob = sample_prob

    def process(self, scenes, tmp_dir):
        stats = RasterStats()
        stats.compute(
            list(map(lambda s: s.raster_source, scenes)),
            sample_prob=self.sample_prob)
        stats.save(self.stats_uri)



# flake8: noqa

from rastervision.task.task import *
from rastervision.task.task_config import *
from rastervision.task.object_detection import *
from rastervision.task.object_detection_config import *
from rastervision.task.chip_classification import *
from rastervision.task.chip_classification_config import *
from rastervision.task.semantic_segmentation import *
from rastervision.task.semantic_segmentation_config import *



from abc import ABC, abstractmethod


class LabelSource(ABC):
    """An interface for storage of labels for a scene.

    An LabelSource is a read source of labels for a scene
    that could be backed by a file, a database, an API, etc. The difference
    between LabelSources and Labels can be understood by analogy to the
    difference between a database and result sets queried from a database.
    """

    @abstractmethod
    def get_labels(self, window=None):
        """Return labels overlapping with window.

        Args:
            window: Box

        Returns:
            Labels overlapping with window. If window is None,
                returns all labels.
        """
        pass



import torch


def compute_conf_mat_metrics(conf_mat):
    # eps is to avoid dividing by zero.
    eps = torch.tensor(1e-4)
    gt_count = conf_mat.sum(dim=1)
    pred_count = conf_mat.sum(dim=0)
    total = conf_mat.sum()
    true_pos = torch.diag(conf_mat)
    precision = true_pos / torch.max(pred_count, eps)
    recall = true_pos / torch.max(gt_count, eps)

    weights = gt_count / total
    weighted_precision = (weights * precision).sum()
    weighted_recall = (weights * recall).sum()
    weighted_f1 = ((2 * weighted_precision * weighted_recall) / torch.max(
        weighted_precision + weighted_recall, eps))
    metrics = {
        'precision': weighted_precision.item(),
        'recall': weighted_recall.item(),
        'f1': weighted_f1.item()
    }
    return metrics



import logging
import math
import os
import pyproj
import subprocess
from decimal import Decimal
import tempfile

import numpy as np
import rasterio
from rasterio.enums import (ColorInterp, MaskFlags)

from rastervision.core.box import Box
from rastervision.data.crs_transformer import RasterioCRSTransformer
from rastervision.utils.files import download_if_needed
from rastervision.data import (ActivateMixin, ActivationError)
from rastervision.data.raster_source import RasterSource

log = logging.getLogger(__name__)
wgs84 = pyproj.Proj({'init': 'epsg:4326'})
wgs84_proj4 = '+init=epsg:4326'
meters_per_degree = 111319.5


def build_vrt(vrt_path, image_paths):
    """Build a VRT for a set of TIFF files."""
    cmd = ['gdalbuildvrt', vrt_path]
    cmd.extend(image_paths)
    subprocess.run(cmd)


def download_and_build_vrt(image_uris, temp_dir):
    log.info('Building VRT...')
    image_paths = [download_if_needed(uri, temp_dir) for uri in image_uris]
    image_path = os.path.join(temp_dir, 'index.vrt')
    build_vrt(image_path, image_paths)
    return image_path


def load_window(image_dataset, window=None, is_masked=False):
    """Load a window of an image using Rasterio.

    Args:
        image_dataset: a Rasterio dataset
        window: ((row_start, row_stop), (col_start, col_stop)) or
        ((y_min, y_max), (x_min, x_max))
        is_masked: If True, read a masked array from rasterio

    Returns:
        np.ndarray of shape (height, width, channels) where channels is the number of
            channels in the image_dataset.
    """
    if is_masked:
        im = image_dataset.read(window=window, boundless=True, masked=True)
        im = np.ma.filled(im, fill_value=0)
    else:
        im = image_dataset.read(window=window, boundless=True)

    # Handle non-zero NODATA values by setting the data to 0.
    for channel, nodata in enumerate(image_dataset.nodatavals):
        if nodata is not None and nodata != 0:
            im[channel, im[channel] == nodata] = 0

    im = np.transpose(im, axes=[1, 2, 0])
    return im


class RasterioSource(ActivateMixin, RasterSource):
    def __init__(self,
                 uris,
                 raster_transformers,
                 temp_dir,
                 channel_order=None,
                 x_shift_meters=0.0,
                 y_shift_meters=0.0):
        """Constructor.

        This RasterSource can read any file that can be opened by Rasterio/GDAL
        including georeferenced formats such as GeoTIFF and non-georeferenced formats
        such as JPG. See https://www.gdal.org/formats_list.html for more details.

        If channel_order is None, then use non-alpha channels. This also sets any
        masked or NODATA pixel values to be zeros.

        Args:
            channel_order: list of indices of channels to extract from raw imagery
        """
        self.uris = uris
        self.temp_dir = temp_dir
        self.image_temp_dir = None
        self.image_dataset = None
        self.x_shift_meters = x_shift_meters
        self.y_shift_meters = y_shift_meters

        num_channels = None

        # Activate in order to get information out of the raster
        with self.activate():
            num_channels = self.image_dataset.count
            if channel_order is None:
                colorinterp = self.image_dataset.colorinterp
                if colorinterp:
                    channel_order = [
                        i for i, color_interp in enumerate(colorinterp)
                        if color_interp != ColorInterp.alpha
                    ]
                else:
                    channel_order = list(range(0, num_channels))
            self.validate_channel_order(channel_order, num_channels)

            mask_flags = self.image_dataset.mask_flag_enums
            self.is_masked = any(
                [m for m in mask_flags if m != MaskFlags.all_valid])

            self.height = self.image_dataset.height
            self.width = self.image_dataset.width

            # Get 1x1 chip and apply raster transformers to test dtype.
            test_chip = self.get_raw_chip(Box.make_square(0, 0, 1))
            test_chip = test_chip[:, :, channel_order]
            for transformer in raster_transformers:
                test_chip = transformer.transform(test_chip, channel_order)
            self.dtype = test_chip.dtype

            self._set_crs_transformer()

        super().__init__(channel_order, num_channels, raster_transformers)

    def _download_data(self, temp_dir):
        """Download any data needed for this Raster Source.

        Return a single local path representing the image or a VRT of the data.
        """
        if len(self.uris) == 1:
            return download_if_needed(self.uris[0], temp_dir)
        else:
            return download_and_build_vrt(self.uris, temp_dir)

    def get_crs_transformer(self):
        return self.crs_transformer

    def get_extent(self):
        return Box(0, 0, self.height, self.width)

    def get_dtype(self):
        """Return the numpy.dtype of this scene"""
        return self.dtype

    def _get_chip(self, window):
        if self.image_dataset is None:
            raise ActivationError('RasterSource must be activated before use')
        shifted_window = self._get_shifted_window(window)
        return load_window(
            self.image_dataset,
            window=shifted_window.rasterio_format(),
            is_masked=self.is_masked)

    def _activate(self):
        # Download images to temporary directory and delete it when done.
        self.image_temp_dir = tempfile.TemporaryDirectory(dir=self.temp_dir)
        self.imagery_path = self._download_data(self.image_temp_dir.name)
        self.image_dataset = rasterio.open(self.imagery_path)
        self._set_crs_transformer()

    def _set_crs_transformer(self):
        self.crs_transformer = RasterioCRSTransformer.from_dataset(
            self.image_dataset)
        self.crs = self.image_dataset.crs
        if self.crs:
            self.proj = pyproj.Proj(self.crs)
        else:
            self.proj = None
        self.crs = str(self.crs)

    def _deactivate(self):
        self.image_dataset.close()
        self.image_dataset = None
        self.image_temp_dir.cleanup()
        self.image_temp_dir = None

    def _get_shifted_window(self, window):
        do_shift = self.x_shift_meters != 0.0 or self.y_shift_meters != 0.0
        if do_shift:
            ymin, xmin, ymax, xmax = window.tuple_format()
            width = window.get_width()
            height = window.get_height()

            # Transform image coordinates into world coordinates
            transform = self.image_dataset.transform
            xmin2, ymin2 = transform * (xmin, ymin)

            # Transform from world coordinates to WGS84
            if self.crs != wgs84_proj4 and self.proj:
                lon, lat = pyproj.transform(self.proj, wgs84, xmin2, ymin2)
            else:
                lon, lat = xmin2, ymin2

            # Shift.  This is performed by computing the shifts in
            # meters to shifts in degrees.  Those shifts are then
            # applied to the WGS84 coordinate.
            #
            # Courtesy of https://gis.stackexchange.com/questions/2951/algorithm-for-offsetting-a-latitude-longitude-by-some-amount-of-meters  # noqa
            lat_radians = math.pi * lat / 180.0
            dlon = Decimal(self.x_shift_meters) / Decimal(
                meters_per_degree * math.cos(lat_radians))
            dlat = Decimal(self.y_shift_meters) / Decimal(meters_per_degree)
            lon = float(Decimal(lon) + dlon)
            lat = float(Decimal(lat) + dlat)

            # Transform from WGS84 to world coordinates
            if self.crs != wgs84_proj4 and self.proj:
                xmin3, ymin3 = pyproj.transform(wgs84, self.proj, lon, lat)
                xmin3 = int(round(xmin3))
                ymin3 = int(round(ymin3))
            else:
                xmin3, ymin3 = lon, lat

            # Trasnform from world coordinates back into image coordinates
            xmin4, ymin4 = ~transform * (xmin3, ymin3)

            window = Box(ymin4, xmin4, ymin4 + height, xmin4 + width)
        return window



# flake8: noqa

# Registry keys

EVALUATOR = 'EVALUATOR'

OBJECT_DETECTION_EVALUATOR = 'OBJECT_DETECTION_EVALUATOR'
CHIP_CLASSIFICATION_EVALUATOR = 'CHIP_CLASSIFICATION_EVALUATOR'
SEMANTIC_SEGMENTATION_EVALUATOR = 'SEMANTIC_SEGMENTATION_EVALUATOR'

from rastervision.evaluation.evaluator_config import EvaluatorConfig



from abc import (ABC, abstractmethod)
from datetime import datetime

import rastervision as rv


class NotReadableError(Exception):
    pass


class NotWritableError(Exception):
    pass


class ProtobufParseException(Exception):
    pass


class FileSystem(ABC):
    @staticmethod
    def get_file_system(uri, mode='r'):
        return rv._registry.get_file_system(uri, mode)

    @staticmethod
    @abstractmethod
    def matches_uri(uri: str, mode: str) -> bool:
        """Returns True if this FileSystem should be used
        for the given URI under the given mode.

        Mode can be 'r' or 'w'
        """
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def file_exists(uri: str, include_dir: bool = True) -> bool:
        """Check if a  file exists.
        Args:
          uri: The URI to check
          include_dir: Include directories in check, if this filesystem
                       supports directory reads. Otherwise only
                       return true if a single file exists at the URI.
        """
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def read_str(uri: str) -> str:
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def read_bytes(uri: str) -> bytes:
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def write_str(uri: str, data: str) -> None:
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def write_bytes(uri: str, data: bytes) -> None:
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def sync_to_dir(src_dir_uri: str, dest_dir_uri: str,
                    delete: bool = False) -> None:
        """Syncs a local directory to a destination.

        Arguments:
           - src_dir_uri: Source directory to sync from. Must be a local directoy.
           - dest_dir_uri: A destination that can be sync to by this FileSystem.
           - delete: True if the destination should be deleted first. Defaults to False.
        """
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def sync_from_dir(src_dir_uri: str,
                      dest_dir_uri: str,
                      delete: bool = False) -> None:
        """Syncs a local directory to a destination.

        Arguments:
           - src_dir_uri: Source directory to sync from. Must be a local directoy.
           - dest_dir_uri: A destination that can be sync to by this FileSystem.
           - delete: True if the destination should be deleted first. Defaults to False.
        """
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def copy_to(src_path: str, dst_uri: str) -> None:
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def copy_from(uri: str, path: str) -> None:
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def local_path(uri: str, download_dir: str) -> None:
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def last_modified(uri: str) -> datetime:
        """Returns the last modified  date in UTC of this URI,
        or None if this FileSystem does not support this operation.
        """
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def list_paths(uri, ext=None):
        pass  # pragma: no cover



from abc import abstractmethod

import numpy as np
import logging

from rastervision.core.training_data import TrainingData

# TODO: DRY... same keys as in ml_backends/tf_object_detection_api.py
TRAIN = 'train'
VALIDATION = 'validation'

log = logging.getLogger(__name__)


class Task(object):
    """Functionality for a specific machine learning task.

    This should be subclassed to add a new task, such as object detection
    """

    def __init__(self, task_config, backend):
        """Construct a new Task.

        Args:
            task_config: TaskConfig
            backend: Backend
        """
        self.config = task_config
        self.backend = backend

    @abstractmethod
    def get_train_windows(self, scene):
        """Return the training windows for a Scene.

        The training windows represent the spatial extent of the training
        chips to generate.

        Args:
            scene: Scene to generate windows for

        Returns:
            list of Boxes
        """
        pass

    @abstractmethod
    def get_train_labels(self, window, scene):
        """Return the training labels in a window for a scene.

        Args:
            window: Box
            scene: Scene

        Returns:
            Labels that lie within window
        """
        pass

    @abstractmethod
    def post_process_predictions(self, labels, scene):
        """Runs a post-processing step on labels at end of prediction.

        Returns:
            Labels
        """
        pass

    @abstractmethod
    def get_predict_windows(self, extent):
        """Return windows to compute predictions for.

        Args:
            extent: Box representing extent of RasterSource

        Returns:
            list of Boxes
        """
        pass

    @abstractmethod
    def save_debug_predict_image(self, scene, debug_dir_uri):
        """Save a debug image of predictions.

        This writes to debug_dir_uri/<scene.id>.jpg.
        """
        pass

    def make_chips(self, train_scenes, validation_scenes, augmentors, tmp_dir):
        """Make training chips.

        Convert Scenes with a ground_truth_label_store into training
        chips in MLBackend-specific format, and write to URI specified in
        options.

        Args:
            train_scenes: list of Scenes
            validation_scenes: list of Scenes
                (that is disjoint from train_scenes)
            augmentors: Augmentors used to augment training data
        """

        def _process_scene(scene, type_, augment):
            with scene.activate():
                data = TrainingData()
                log.info('Making {} chips for scene: {}'.format(
                    type_, scene.id))
                windows = self.get_train_windows(scene)
                for window in windows:
                    chip = scene.raster_source.get_chip(window)
                    labels = self.get_train_labels(window, scene)
                    data.append(chip, window, labels)
                # Shuffle data so the first N samples which are displayed in
                # Tensorboard are more diverse.
                data.shuffle()

                # Process augmentation
                if augment:
                    for augmentor in augmentors:
                        data = augmentor.process(data, tmp_dir)

                return self.backend.process_scene_data(scene, data, tmp_dir)

        def _process_scenes(scenes, type_, augment):
            return [_process_scene(scene, type_, augment) for scene in scenes]

        processed_training_results = _process_scenes(
            train_scenes, TRAIN, augment=True)
        processed_validation_results = _process_scenes(
            validation_scenes, VALIDATION, augment=False)

        self.backend.process_sceneset_results(
            processed_training_results, processed_validation_results, tmp_dir)

    def train(self, tmp_dir):
        """Train a model.
        """
        self.backend.train(tmp_dir)

    def predict(self, scenes, tmp_dir):
        """Make predictions for scenes.

        The predictions are saved to the prediction_label_store in
        each scene.

        Args:
            scenes: list of Scenes
        """
        self.backend.load_model(tmp_dir)

        for scene in scenes:
            with scene.activate():
                labels = self.predict_scene(scene, tmp_dir)
                label_store = scene.prediction_label_store
                label_store.save(labels)

                if self.config.debug and self.config.predict_debug_uri:
                    self.save_debug_predict_image(
                        scene, self.config.predict_debug_uri)

    def predict_scene(self, scene, tmp_dir):
        """Predict on a single scene, and return the labels."""
        log.info('Making predictions for scene')
        raster_source = scene.raster_source
        label_store = scene.prediction_label_store
        labels = label_store.empty_labels()

        windows = self.get_predict_windows(raster_source.get_extent())

        def predict_batch(predict_chips, predict_windows):
            nonlocal labels
            new_labels = self.backend.predict(
                np.array(predict_chips), predict_windows, tmp_dir)
            labels += new_labels
            print('.' * len(predict_chips), end='', flush=True)

        batch_chips, batch_windows = [], []
        for window in windows:
            chip = raster_source.get_chip(window)
            if np.any(chip):
                batch_chips.append(chip)
                batch_windows.append(window)

            # Predict on batch
            if len(batch_chips) >= self.config.predict_batch_size:
                predict_batch(batch_chips, batch_windows)
                batch_chips, batch_windows = [], []
        print()

        # Predict on remaining batch
        if len(batch_chips) > 0:
            predict_batch(batch_chips, batch_windows)

        return self.post_process_predictions(labels, scene)



import networkx as nx
import logging

import rastervision as rv
from rastervision.utils.files import file_exists

import click

log = logging.getLogger(__name__)


class CommandDAG:
    """ A directed acyclic graph of command definitions.
    """

    def __init__(self,
                 command_definitions,
                 rerun_commands=False,
                 skip_file_check=False):
        """Generates a CommandDAG from a list of CommandDefinitions

        This logic checks if there are any non-exsiting URIs that are
        not produced as outputs by some command in the set. If so,
        it raises a ConfigError stating the missing files.
        """
        # Create a set of edges, from input_uri to command_config and
        # from command_config to output_uri. Nodes for commands are their
        # index into command_definitions.

        uri_dag = nx.DiGraph()

        log.debug('Creating command and URI DAG from command definitions...')
        for idx, command_def in enumerate(command_definitions):
            uri_dag.add_node(idx)
            for input_uri in command_def.io_def.input_uris:
                uri_dag.add_edge(input_uri, idx)

            for output_uri in command_def.io_def.output_uris:
                uri_dag.add_edge(idx, output_uri)

        # Find all source input_uris, and ensure they exist.
        if not skip_file_check:
            log.debug('Ensuring input files exist...')
            unsolved_sources = [
                uri for uri in uri_dag.nodes
                if (type(uri) == str and len(uri_dag.in_edges(uri)) == 0)
            ]

            missing_files = []

            with click.progressbar(
                    unsolved_sources,
                    label='Ensuring input files exist  ') as uris:
                for uri in uris:
                    if not file_exists(uri):
                        missing_files.append(uri)

            if any(missing_files):
                raise rv.ConfigError(
                    'Files do not exist and are not supplied by commands:\n'
                    '\t{}\n'.format(',\b\t'.join(missing_files)))

        # If we are not rerunning, remove commands that have existing outputs.
        self.skipped_commands = []
        if not rerun_commands:
            log.debug('Checking for existing output...')
            commands_to_outputs = [(idx, edge[1]) for idx in uri_dag.nodes
                                   if type(idx) == int
                                   for edge in uri_dag.out_edges(idx)]
            with click.progressbar(
                    commands_to_outputs,
                    label='Checking for existing output') as lst:
                for idx, output_uri in lst:
                    if file_exists(output_uri):
                        uri_dag.remove_edge(idx, output_uri)

            for idx in set(map(lambda x: x[0], commands_to_outputs)):
                if len(uri_dag.out_edges(idx)) == 0:
                    self.skipped_commands.append(command_definitions[idx])
                    uri_dag.remove_node(idx)

        # Collapse the graph to create edges from command to command.
        command_id_dag = nx.DiGraph()

        log.debug('Creating DAG of commands...')
        for idx in [idx for idx in uri_dag.nodes if (type(idx) == int)]:
            command_id_dag.add_node(idx)
            for upstream_idx in [
                    edge2[0] for edge1 in uri_dag.in_edges(idx)
                    for edge2 in uri_dag.in_edges(edge1[0])
            ]:
                command_id_dag.add_edge(upstream_idx, idx)

        # Feed this digraph of commands to the child runner.
        self.command_definitions = command_definitions
        self.command_id_dag = command_id_dag

    def get_sorted_commands(self):
        """Return a topologically sorted list of commands configurations.

        Returns a list of command configurations that are sorted such that every
        command that depends on some other parent command appears later
        than that parent command.
        """
        return [
            self.command_definitions[idx].command_config
            for idx in self.get_sorted_command_ids()
        ]

    def get_sorted_command_ids(self):
        """Return a topologically sorted list of commands ids.

        Returns a list of command IDs that can be used to retrieve
        specific commands out of this DAG. These are sorted such that every
        command that depends on some other parent command appears later
        than that parent command.
        """
        return [idx for idx in nx.topological_sort(self.command_id_dag)]

    def get_command(self, command_id):
        """Retrieves a command configuration for the given ID"""
        return self.get_command_definition(command_id).command_config

    def get_command_definition(self, command_id):
        """Retrieves a command definition for the given ID"""
        return self.command_definitions[command_id]

    def get_upstream_command_ids(self, command_id):
        """Returns the command ids for upstream commands for the command
        with the given id.
        """
        return list(
            map(lambda x: x[0], self.command_id_dag.in_edges(command_id)))

    def get_command_definitions(self):
        """Returns the command definitions that will be run in this DAG."""
        return [
            self.command_definitions[idx] for idx in self.command_id_dag.nodes
        ]



# flake8: noqa

# Registry Keys

TASK = 'TASK'

OBJECT_DETECTION = 'OBJECT_DETECTION'
CHIP_CLASSIFICATION = 'CHIP_CLASSIFICATION'
SEMANTIC_SEGMENTATION = 'SEMANTIC_SEGMENTATION'

from .task_config import TaskConfig



# flake8: noqa

from rastervision.data.raster_source.raster_source import *
from rastervision.data.raster_source.raster_source_config import *
from rastervision.data.raster_source.rasterio_source import *
from rastervision.data.raster_source.rasterio_source_config import *
from rastervision.data.raster_source.rasterized_source import *
from rastervision.data.raster_source.rasterized_source_config import *



import matplotlib.patches as patches


def plot_xy(ax, x, y=None, label_names=None):
    ax.imshow(x.permute(1, 2, 0))

    if y is not None:
        scores = y.get_field('scores')
        for box_ind, (box, label) in enumerate(
                zip(y.boxes, y.get_field('labels'))):
            rect = patches.Rectangle(
                (box[1], box[0]),
                box[3] - box[1],
                box[2] - box[0],
                linewidth=1,
                edgecolor='cyan',
                facecolor='none')
            ax.add_patch(rect)

            label_name = label_names[label]
            if scores is not None:
                score = scores[box_ind]
                label_name += ' {:.2f}'.format(score)

            h, w = x.shape[1:]
            label_height = h * 0.03
            label_width = w * 0.15
            rect = patches.Rectangle(
                (box[1], box[0] - label_height),
                label_width,
                label_height,
                color='cyan')
            ax.add_patch(rect)

            ax.text(
                box[1] + w * 0.003, box[0] - h * 0.003, label_name, fontsize=7)

    ax.axis('off')



import torch
import torch.nn as nn
from torchvision.models.detection.faster_rcnn import FasterRCNN
from torchvision.models.detection.backbone_utils import BackboneWithFPN
from torchvision.models import resnet
from torchvision.ops import misc as misc_nn_ops

from rastervision.backend.torch_utils.object_detection.boxlist import BoxList


def get_out_channels(model):
    out = {}

    def make_save_output(layer_name):
        def save_output(layer, input, output):
            out[layer_name] = output.shape[1]

        return save_output

    model.layer1.register_forward_hook(make_save_output('layer1'))
    model.layer2.register_forward_hook(make_save_output('layer2'))
    model.layer3.register_forward_hook(make_save_output('layer3'))
    model.layer4.register_forward_hook(make_save_output('layer4'))

    model(torch.empty((1, 3, 128, 128)))
    return [out['layer1'], out['layer2'], out['layer3'], out['layer4']]


# This fixes a bug in torchvision.
def resnet_fpn_backbone(backbone_name, pretrained):
    backbone = resnet.__dict__[backbone_name](
        pretrained=pretrained, norm_layer=misc_nn_ops.FrozenBatchNorm2d)

    # freeze layers
    for name, parameter in backbone.named_parameters():
        if 'layer2' not in name and 'layer3' not in name and 'layer4' not in name:
            parameter.requires_grad_(False)

    return_layers = {'layer1': 0, 'layer2': 1, 'layer3': 2, 'layer4': 3}

    out_channels = 256
    in_channels_list = get_out_channels(backbone)
    return BackboneWithFPN(backbone, return_layers, in_channels_list,
                           out_channels)


class MyFasterRCNN(nn.Module):
    """Adapter around torchvision Faster-RCNN.

    The purpose of the adapter is to use a different input and output format
    and inject bogus boxes to circumvent torchvision's inability to handle
    training examples with no ground truth boxes.
    """

    def __init__(self, backbone_arch, num_labels, img_sz, pretrained=True):
        super().__init__()

        backbone = resnet_fpn_backbone(backbone_arch, pretrained)
        self.model = FasterRCNN(
            backbone, num_labels, min_size=img_sz, max_size=img_sz)
        self.subloss_names = [
            'total_loss', 'loss_box_reg', 'loss_classifier', 'loss_objectness',
            'loss_rpn_box_reg'
        ]

    def forward(self, input, targets=None):
        """Forward pass

        Args:
            input: tensor<n, 3, h, w> with batch of images
            targets: None or list<BoxList> of length n with boxes and labels

        Returns:
            if targets is None, returns list<BoxList> of length n, containing
            boxes, labels, and scores for boxes with score > 0.05. Further
            filtering based on score should be done before considering the
            prediction "final".

            if targets is a list, returns the losses as dict with keys from
            self.subloss_names.
        """
        if targets:
            # Add bogus background class box for each image to workaround
            # the inability of torchvision to train on images with
            # no ground truth boxes. This is important for being able
            # to handle negative chips generated by RV.
            new_targets = []
            for x, y in zip(input, targets):
                h, w = x.shape[1:]
                boxes = torch.cat(
                    [
                        y.boxes,
                        torch.tensor([[0., 0, h, w]], device=input.device)
                    ],
                    dim=0)
                labels = torch.cat(
                    [
                        y.get_field('labels'),
                        torch.tensor([0], device=input.device)
                    ],
                    dim=0)
                bl = BoxList(boxes, labels=labels)
                new_targets.append(bl)
            targets = new_targets

            _targets = [bl.xyxy() for bl in targets]
            _targets = [{
                'boxes': bl.boxes,
                'labels': bl.get_field('labels')
            } for bl in _targets]
            loss_dict = self.model(input, _targets)
            loss_dict['total_loss'] = sum(list(loss_dict.values()))
            return loss_dict

        out = self.model(input)
        boxlists = [
            BoxList(
                _out['boxes'], labels=_out['labels'],
                scores=_out['scores']).yxyx() for _out in out
        ]

        # Remove bogus background boxes.
        new_boxlists = []
        for bl in boxlists:
            labels = bl.get_field('labels')
            non_zero_inds = labels != 0
            new_boxlists.append(bl.ind_filter(non_zero_inds))
        return new_boxlists



from copy import deepcopy
from typing import Union

from shapely.geometry import shape

import rastervision as rv
from rastervision.core import (Config, ConfigBuilder, BundledConfigMixin)
from rastervision.task import TaskConfig
from rastervision.data import (Scene, RasterSourceConfig, LabelSourceConfig,
                               LabelStoreConfig, GeoJSONVectorSource)
from rastervision.protos.scene_pb2 \
    import SceneConfig as SceneConfigMsg


class SceneConfig(BundledConfigMixin, Config):
    def __init__(self,
                 id,
                 raster_source,
                 label_source=None,
                 label_store=None,
                 aoi_uris=None):
        self.id = id
        self.raster_source = raster_source
        self.label_source = label_source
        self.label_store = label_store
        self.aoi_uris = aoi_uris

    def create_scene(self, task_config: TaskConfig, tmp_dir: str) -> Scene:
        """Create this scene.

           Args:
              task - TaskConfig
              tmp_dir - Temporary directory to use
        """
        raster_source = self.raster_source.create_source(tmp_dir)
        extent = raster_source.get_extent()
        crs_transformer = raster_source.get_crs_transformer()

        label_source = None
        if self.label_source:
            label_source = self.label_source.create_source(
                task_config, extent, crs_transformer, tmp_dir)
        label_store = None
        if self.label_store:
            label_store = self.label_store.create_store(
                task_config, extent, crs_transformer, tmp_dir)
        aoi_polygons = None
        if self.aoi_uris:
            aoi_polygons = []
            for uri in self.aoi_uris:
                aoi_geojson = GeoJSONVectorSource(
                    uri, crs_transformer).get_geojson()
                for f in aoi_geojson['features']:
                    aoi_polygons.append(shape(f['geometry']))

        return Scene(self.id, raster_source, label_source, label_store,
                     aoi_polygons)

    def to_proto(self):
        msg = SceneConfigMsg(
            id=self.id,
            raster_source=self.raster_source.to_proto(),
            aoi_uris=self.aoi_uris)

        if self.label_source:
            msg.ground_truth_label_source.CopyFrom(
                self.label_source.to_proto())
        if self.label_store:
            msg.prediction_label_store.CopyFrom(self.label_store.to_proto())
        return msg

    def save_bundle_files(self, bundle_dir):
        new_source, files = self.raster_source.save_bundle_files(bundle_dir)
        new_config = self.to_builder() \
                         .with_raster_source(new_source) \
                         .build()
        return (new_config, files)

    def load_bundle_files(self, bundle_dir):
        new_source = self.raster_source.load_bundle_files(bundle_dir)
        return self.to_builder() \
                   .with_raster_source(new_source) \
                   .build()

    def for_prediction(self, image_uri, label_uri=None):
        """Creates a version of this scene that is set to
        predict against the image_uri. If label_uri is set,
        the scene must already have a label_store.
        """
        new_source = self.raster_source.for_prediction(image_uri)
        b = self.to_builder().with_raster_source(new_source)

        if label_uri:
            if not self.label_store:
                raise rv.ConfigError('Cannot call for_prediciton on  a '
                                     'scene that does not have a label '
                                     'store set.')
            new_store = self.label_store.for_prediction(label_uri)
            b = b.with_label_store(new_store)

        return b.build()

    def create_local(self, tmp_dir):
        new_source = self.raster_source.create_local(tmp_dir)
        return self.to_builder() \
                   .with_raster_source(new_source) \
                   .build()

    def to_builder(self):
        return SceneConfigBuilder(self)

    def update_for_command(self, command_type, experiment_config,
                           context=None):
        if context is None:
            context = []
        context = context + [self]

        self.raster_source.update_for_command(command_type, experiment_config,
                                              context)

        if self.label_source:
            self.label_source.update_for_command(command_type,
                                                 experiment_config, context)

        if self.label_store:
            self.label_store.update_for_command(command_type,
                                                experiment_config, context)

    def report_io(self, command_type, io_def):
        self.raster_source.report_io(command_type, io_def)

        if self.label_source:
            self.label_source.report_io(command_type, io_def)

        if self.label_store:
            self.label_store.report_io(command_type, io_def)

        if self.aoi_uris:
            for uri in self.aoi_uris:
                io_def.add_input(uri)

    @staticmethod
    def builder():
        return SceneConfigBuilder()

    @staticmethod
    def from_proto(msg):
        """Creates a SceneConfig from the specificed protobuf message
        """
        return SceneConfigBuilder().from_proto(msg).build()


class SceneConfigBuilder(ConfigBuilder):
    def __init__(self, prev=None):
        config = {}
        if prev:
            config = {
                'id': prev.id,
                'raster_source': prev.raster_source,
                'label_source': prev.label_source,
                'label_store': prev.label_store,
                'aoi_uris': prev.aoi_uris
            }
        super().__init__(SceneConfig, config)
        self.task = None

    def from_proto(self, msg):
        b = self.with_id(msg.id) \
                .with_raster_source(RasterSourceConfig.from_proto(msg.raster_source))
        if msg.HasField('ground_truth_label_source'):
            b = b.with_label_source(
                LabelSourceConfig.from_proto(msg.ground_truth_label_source))
        if msg.HasField('prediction_label_store'):
            b = b.with_label_store(
                LabelStoreConfig.from_proto(msg.prediction_label_store))

        # Here for backward compatibility.
        if msg.HasField('aoi_uri'):
            b = b.with_aoi_uri(msg.aoi_uri)

        if len(msg.aoi_uris):
            b = b.with_aoi_uris(msg.aoi_uris)

        return b

    def with_task(self, task):
        """Sets a specific task type, e.g. rv.OBJECT_DETECTION."""
        b = deepcopy(self)
        b.task = task
        return b

    def with_id(self, id):
        """Sets an id for the scene."""
        b = deepcopy(self)
        b.config['id'] = id
        return b

    def with_raster_source(self,
                           raster_source: Union[str, RasterSourceConfig],
                           channel_order=None):
        """
        Sets the raster source for this scene.

        Args:
            raster_source: Can either be a raster source configuration, or
                a string. If a string, the registry will be queried
                to grab the default RasterSourceConfig for the string.
            channel_order: Optional channel order for this raster source.
        """
        b = deepcopy(self)
        if isinstance(raster_source, RasterSourceConfig):
            if channel_order is not None:
                rs = raster_source.to_builder() \
                                  .with_channel_order(channel_order) \
                                  .build()
                b.config['raster_source'] = rs
            else:
                b.config['raster_source'] = raster_source
        else:
            provider = rv._registry.get_raster_source_default_provider(
                raster_source)
            b.config['raster_source'] = provider.construct(
                raster_source, channel_order)

        return b

    def with_label_source(self, label_source: Union[str, LabelSourceConfig]):
        """
        Sets the raster source for this scene.

        Args:
           label_source: Can either be a label source configuration, or
                         a string. If a string, the registry will be queried
                         to grab the default LabelSourceConfig for the string.

        Note:
           A task must be set with `with_task` before calling this, if calling
           with a string.
        """
        b = deepcopy(self)
        if isinstance(label_source, LabelSourceConfig):
            b.config['label_source'] = label_source
        else:
            if not self.task:
                raise rv.ConfigError(
                    "You must set a task with '.with_task' before "
                    'creating a default label store for {}'.format(
                        label_source))
            provider = rv._registry.get_label_source_default_provider(
                self.task.task_type, label_source)
            b.config['label_source'] = provider.construct(label_source)

        return b

    def clear_label_source(self):
        """Clears the label source for this scene"""
        b = deepcopy(self)
        b.config['label_source'] = None
        return b

    def with_label_store(
            self, label_store: Union[str, LabelStoreConfig, None] = None):
        """
        Sets the raster store for this scene.

        Args:
           label_store: Can either be a label store configuration, or
                        a string, or None. If a string, the registry will
                        be queried to grab the default LabelStoreConfig for
                        the string. If None, then the default for the task
                        from the regsitry will be used.

        Note:
           A task must be set with `with_task` before calling this, if calling
           with a string.
        """
        b = deepcopy(self)
        if isinstance(label_store, LabelStoreConfig):
            b.config['label_store'] = label_store
        elif isinstance(label_store, str):
            if not self.task:
                raise rv.ConfigError(
                    "You must set a task with '.with_task' before "
                    'creating a default label store for {}'.format(
                        label_store))
            provider = rv._registry.get_label_store_default_provider(
                self.task.task_type, label_store)
            b.config['label_store'] = provider.construct(label_store)
        else:
            if not self.task:
                raise rv.ConfigError(
                    "You must set a task with '.with_task' before "
                    'creating a default label store.')
            provider = rv._registry.get_label_store_default_provider(
                self.task.task_type)
            b.config['label_store'] = provider.construct()

        return b

    def clear_label_store(self):
        """Clears the label store for this scene"""
        b = deepcopy(self)
        b.config['label_store'] = None
        return b

    def with_aoi_uri(self, uri):
        """Sets the Area of Interest for the scene.

            Args:
                uri: a URI of a GeoJSON file with polygons.

        """
        b = deepcopy(self)
        b.config['aoi_uris'] = [uri]
        return b

    def with_aoi_uris(self, uris):
        """Sets the areas of interest for the scene.

            Args:
                uris: List of URIs each to a GeoJSON file with polygons.
        """
        b = deepcopy(self)
        b.config['aoi_uris'] = uris
        return b

    def clear_aois(self):
        """Clears the AOIs for this scene"""
        b = deepcopy(self)
        b.config['aoi_uris'] = None
        return b



import os
import glob
import shutil
import tarfile
import uuid
from typing import (Dict, List, Tuple)
from os.path import join
from subprocess import Popen
import logging

import numpy as np
from google.protobuf import (json_format)

import rastervision as rv
from rastervision.core.box import Box
from rastervision.core.class_map import ClassMap
from rastervision.backend import Backend
from rastervision.data.scene import Scene
from rastervision.data.label import SemanticSegmentationLabels
from rastervision.core.training_data import TrainingData
from rastervision.backend.tf_object_detection import (write_tf_record, TRAIN,
                                                      VALIDATION)
from rastervision.protos.deeplab.train_pb2 import (TrainingParameters as
                                                   TrainingParametersMsg)
from rastervision.utils.files import (
    download_if_needed, get_local_path, make_dir, start_sync, upload_or_copy,
    sync_to_dir, sync_from_dir, list_paths, file_exists)
from rastervision.utils.misc import (numpy_to_png, png_to_numpy, save_img,
                                     terminate_at_exit)
from rastervision.data.label_source.utils import color_to_integer
from rastervision.rv_config import RVConfig

FROZEN_INFERENCE_GRAPH = 'model'
INPUT_TENSOR_NAME = 'ImageTensor:0'
OUTPUT_TENSOR_NAME = 'SemanticPredictions:0'

log = logging.getLogger(__name__)


def make_tf_examples(training_data: TrainingData, class_map: ClassMap) -> List:
    """Take training data and a class map and return a list of TFRecords.

    Args:
         training_data: A rastervision.core.training_data.TrainingData
              object.
         class_map: A rastervision.core.class_map.ClassMap object.

    Returns:
         list(tensorflow.core.example.example_pb2.Example)

    """
    tf_examples = []
    log.info('Creating TFRecord')
    for chip, window, labels in training_data:
        tf_example = create_tf_example(chip, window,
                                       labels.get_label_arr(window), class_map)
        tf_examples.append(tf_example)
    return tf_examples


def merge_tf_records(output_path: str, src_records: List[str]) -> None:
    """Merge multiple TFRecord files into one.

    Args:
         output_path: Where to write the merged TFRecord file.
         src_records: A list of strings giving the location of the
              input TFRecord files.

    Returns:
         None

    """
    import tensorflow as tf

    records = 0
    with tf.python_io.TFRecordWriter(output_path) as writer:
        log.info('Merging TFRecords')
        for src_record in src_records:
            for string_record in tf.python_io.tf_record_iterator(src_record):
                writer.write(string_record)
                records = records + 1
        log.info('{} records'.format(records))


def make_debug_images(record_path: str, output_dir: str, class_map: ClassMap,
                      p: float) -> None:
    """Render a random sample of the TFRecords in a given file as
    human-viewable PNG files.

    Args:
         record_path: Path to the TFRecord file.
         output_dir: Destination directory for the generated PNG files.
         p: The probability of rendering a particular record.

    Returns:
         None

    """
    import tensorflow as tf
    make_dir(output_dir)

    ids = class_map.get_keys()
    color_strs = list(map(lambda c: c.color, class_map.get_items()))
    color_ints = list(map(lambda c: color_to_integer(c), color_strs))
    correspondence = dict(zip(ids, color_ints))

    def _label_fn(v: int) -> int:
        if v in correspondence:
            return correspondence.get(v)
        else:
            return 0

    label_fn = np.vectorize(_label_fn, otypes=[np.uint64])

    def _image_fn(pixel: int) -> int:
        if (pixel & 0x00ffffff):
            r = ((pixel >> 41 & 0x7f) + (pixel >> 17 & 0x7f)) << 16
            g = ((pixel >> 33 & 0x7f) + (pixel >> 9 & 0x7f)) << 8
            b = ((pixel >> 25 & 0x7f) + (pixel >> 1 & 0x7f)) << 0
            return r + g + b
        else:
            return pixel >> 24

    image_fn = np.vectorize(_image_fn, otypes=[np.uint64])

    log.info('Generating debug chips')
    tfrecord_iter = tf.python_io.tf_record_iterator(record_path)
    for ind, example in enumerate(tfrecord_iter):
        if np.random.rand() <= p:
            example = tf.train.Example.FromString(example)
            im_unpacked, labels = parse_tf_example(example)

            im_r = np.array(im_unpacked[:, :, 0], dtype=np.uint64) * 1 << 40
            im_g = np.array(im_unpacked[:, :, 1], dtype=np.uint64) * 1 << 32
            im_b = np.array(im_unpacked[:, :, 2], dtype=np.uint64) * 1 << 24
            im_packed = im_r + im_g + im_b

            labels_packed = label_fn(np.array(labels))
            im_labels_packed = im_packed + labels_packed
            im_packed = image_fn(im_labels_packed)

            im_unpacked[:, :, 0] = np.bitwise_and(
                im_packed >> 16, 0xff, dtype=np.uint8)
            im_unpacked[:, :, 1] = np.bitwise_and(
                im_packed >> 8, 0xff, dtype=np.uint8)
            im_unpacked[:, :, 2] = np.bitwise_and(
                im_packed >> 0, 0xff, dtype=np.uint8)

            output_path = join(output_dir, '{}.png'.format(ind))
            save_img(im_unpacked, output_path)


def parse_tf_example(example) -> Tuple[np.ndarray, np.ndarray]:
    """Parse a TensorFlow Example into an image array and a label array.

    Args:
         example: A TensorFlow Example object.

    Returns:
         A np.ndarray × np.ndarray pair.

    """
    ie = 'image/encoded'
    isce = 'image/segmentation/class/encoded'
    image_encoded = \
        example.features.feature[ie].bytes_list.value[0]
    image_segmentation_class_encoded = \
        example.features.feature[isce].bytes_list.value[0]
    im = png_to_numpy(image_encoded)
    labels = png_to_numpy(image_segmentation_class_encoded)
    return im, labels


def create_tf_example(image: np.ndarray,
                      window: Box,
                      labels: np.ndarray,
                      class_map: ClassMap,
                      chip_id: str = ''):
    """Create a TensorFlow Example from an image, the labels, &c.

    Args:
         image: An np.ndarray containing the image data.
         window: A Box object containing the bounding box for this example.
         labels: An nd.array containing the label data.
         class_map: A ClassMap object containing mappings between
              numerical and textual labels.
         chip_id: The chip id as a string.

    Returns:
         A DeepLab-compatible TensorFlow Example object containing the
         given data.

    """
    import tensorflow as tf
    from object_detection.utils import dataset_util

    class_keys = set(class_map.get_keys())

    def _clean(n):
        return (n if n in class_keys else 0x00)

    clean = np.vectorize(_clean, otypes=[np.uint8])

    image_encoded = numpy_to_png(image)
    image_filename = chip_id.encode('utf8')
    image_format = 'png'.encode('utf8')
    image_height, image_width, image_channels = image.shape
    image_segmentation_class_encoded = numpy_to_png(clean(labels))
    image_segmentation_class_format = 'png'.encode('utf8')

    features = tf.train.Features(
        feature={
            'image/encoded':
            dataset_util.bytes_feature(image_encoded),
            'image/filename':
            dataset_util.bytes_feature(image_filename),
            'image/format':
            dataset_util.bytes_feature(image_format),
            'image/height':
            dataset_util.int64_feature(image_height),
            'image/width':
            dataset_util.int64_feature(image_width),
            'image/channels':
            dataset_util.int64_feature(image_channels),
            'image/segmentation/class/encoded':
            dataset_util.bytes_feature(image_segmentation_class_encoded),
            'image/segmentation/class/format':
            dataset_util.bytes_feature(image_segmentation_class_format),
        })

    return tf.train.Example(features=features)


def get_record_uri(base_uri: str, split: str, suffix: str) -> str:
    """Given a base URI and a split, return a filename to use.

    Args:
         base_uri: The directory under-which the returned record uri
              will reside.
         split: The split ("train", "validate", et cetera).
         suffix: Unique identifier to place at the end  of the tf record file.
    Returns:
         A uri, under the base_uri, that can be used to store a record
         file.

    """
    return join(base_uri, split, '{}-{}.record'.format(split, suffix))


def get_record_dir(base_uri: str, split: str) -> str:
    """Given a base URI and a split, return the directory URI
    that contains the split records.

    Args:
         base_uri: The directory under-which the returned record uri
              will reside.
         split: The split ("train", "validate", et cetera).
    Returns:
         A uri for a directory, under the base_uri, that contains
    the TF records for the given split.

    """
    return join(base_uri, split)


def get_latest_checkpoint(train_logdir_local: str) -> str:
    """Return the most recently generated checkpoint.

    Args:
         train_logir_local: The directory in-which to look for the
              latest checkpoint.

    Returns:
         Returns the (local) URI to the latest checkpoint.

    """
    ckpts = glob.glob(join(train_logdir_local, 'model.ckpt-*.meta'))
    times = map(os.path.getmtime, ckpts)
    latest = sorted(zip(times, ckpts))[-1][1]
    return latest[:len(latest) - len('.meta')]


def get_evaluation_args(eval_py: str, train_logdir_local: str,
                        dataset_dir_local: str, eval_logdir: str, tfdl_config):
    """Generate the array of arguments needed to run the eval script.

    Args:
         eval_py: The URI of the eval script.
         train_logdir_local: The directory in-which checkpoints can be
              found.
         dataset_dir_local: The directory in which the records are
              found.
         eval_logdir: The directory where evaluation events should be
              logged.
         tfdl_config: google.protobuf.Struct with fields from
            rv.protos.deeplab.train.proto containing TF Deeplab training configuration

    Returns:
         A list of arguments suitable for running the eval script.

    """
    fields = [
        'dataset',
        'output_stride',
        'decoder_output_stride',
        'model_variant',
        'eval_split',
    ]

    multi_fields = [
        'atrous_rates',
        'eval_crop_size',
    ]

    args = ['python', eval_py]

    args.append('--checkpoint_dir={}'.format(train_logdir_local))
    args.append('--eval_logdir={}'.format(eval_logdir))
    args.append('--dataset_dir={}'.format(dataset_dir_local))

    for field in multi_fields:
        for item in tfdl_config.__getattribute__(field):
            args.append('--{}={}'.format(field, item))

    for field in fields:
        field_value = tfdl_config.__getattribute__(field)
        if (not type(field_value) is str) or (not len(field_value) == 0):
            args.append('--{}={}'.format(field, field_value))

    return args


def get_training_args(train_py: str, train_logdir_local: str, tfic_ckpt: str,
                      dataset_dir_local: str, num_classes: int,
                      tfdl_config) -> Tuple[List[str], Dict[str, str]]:
    """Generate the array of arguments needed to run the training script.

    Args:
         train_py: The URI of the training script.
         train_logdir_local: The directory in-which checkpoints will
              be placed.
         tfic_ckpt: URI of the .ckpt "file" from the initial
              checkpoint tarball.
         dataset_dir_local: The directory in which the records are
              found.
         num_classes: The number of classes.
         tfdl_config: google.protobuf.Struct with fields from
            rv.protos.deeplab.train.proto containing TF Deeplab training configuration

    Returns:
         A tuple of two things: (1) a list of arguments suitable for
         starting the training script and (2) an environment in-which
         to start the training script.

    """
    fields = [
        'fine_tune_batch_norm',
        'initialize_last_layer',
        'last_layers_contain_logits_only',
        'save_summaries_images',
        'upsample_logits',
        'base_learning_rate',
        'last_layer_gradient_multiplier',
        'learning_power',
        'learning_rate_decay_factor',
        'max_scale_factor',
        'min_scale_factor',
        'momentum',
        'num_clones',
        'scale_factor_step_size',
        'slow_start_learning_rate',
        'weight_decay',
        'decoder_output_stride',
        'learning_rate_decay_step',
        'output_stride',
        'save_interval_secs',
        'save_summaries_secs',
        'slow_start_step',
        'train_batch_size',
        'training_number_of_steps',
        'dataset',
        'learning_policy',
        'model_variant',
        'train_split',
    ]

    multi_fields = [
        'atrous_rates',
        'train_crop_size',
    ]

    env_fields = [
        'dl_custom_train',
        'dl_custom_validation',
    ]

    args = ['python', train_py]

    args.append('--train_logdir={}'.format(train_logdir_local))
    args.append('--tf_initial_checkpoint={}'.format(tfic_ckpt))
    args.append('--dataset_dir={}'.format(dataset_dir_local))

    for field in multi_fields:
        for item in tfdl_config.__getattribute__(field):
            args.append('--{}={}'.format(field, item))

    for field in fields:
        field_value = tfdl_config.__getattribute__(field)
        if (not type(field_value) is str) or (not len(field_value) == 0):
            args.append('--{}={}'.format(field, field_value))

    env = os.environ.copy()
    for field in env_fields:
        field_value = tfdl_config.__getattribute__(field)
        log.info('{}={}'.format(field.upper(), field_value))
        env[field.upper()] = str(field_value)
    log.info('DL_CUSTOM_CLASSES={}'.format(num_classes))
    env['DL_CUSTOM_CLASSES'] = str(num_classes)

    return (args, env)


def get_export_args(export_py: str, train_logdir_local: str, num_classes: int,
                    tfdl_config) -> List[str]:
    """Generate the array of arguments needed to run the export script.

    Args:
         export_py: The URI of the export script.
         train_logdir_local: The directory in-which checkpoints will
              be placed.
         num_classes: The number of classes.
         tfdl_config: google.protobuf.Struct with fields from
            rv.protos.deeplab.train.proto containing TF Deeplab training configuration

    Returns:
         A list of arguments suitable for starting the training
         script.
    """

    fields = [
        'decoder_output_stride',
        'output_stride',
        'model_variant',
    ]

    args = ['python', export_py]

    args.append('--checkpoint_path={}'.format(
        get_latest_checkpoint(train_logdir_local)))
    args.append('--export_path={}'.format(
        join(train_logdir_local, FROZEN_INFERENCE_GRAPH)))
    args.append('--num_classes={}'.format(num_classes))

    for field in fields:
        field_value = tfdl_config.__getattribute__(field)
        args.append('--{}={}'.format(field, field_value))

    for item in tfdl_config.__getattribute__('atrous_rates'):
        args.append('--{}={}'.format('atrous_rates', item))

    for item in tfdl_config.__getattribute__('train_crop_size'):
        args.append('--{}={}'.format('crop_size', item))

    return args


class TFDeeplab(Backend):
    """Backend-derived type that implements the TensorFlow DeepLab
    backend.

    """

    def __init__(self, backend_config, task_config):
        """Constructor.

        Args:
            backend_config: rv.backend.TFDeeplabConfig
            task_config: rv.task.SemanticSegmentationConfig
        """
        self.sess = None
        self.backend_config = backend_config
        self.task_config = task_config
        self.class_map = task_config.class_map

    def process_scene_data(self, scene: Scene, data: TrainingData,
                           tmp_dir: str) -> str:
        """Process the given scene and data into a TFRecord file specifically
        associated with that file.

        Args:
             scene: The scene data (labels stores, the raster sources,
                  and so on).
             data: The training data.
             tmp_dir: (str) temporary directory to use
        Returns:
            The local path to the generated file.
        """
        # Currently TF Deeplab can only handle uint8
        if scene.raster_source.get_dtype() != np.uint8:
            raise Exception('Cannot use {} backend for imagery that does '
                            'not have data type uint8. '
                            'Use the StatsAnalyzer and StatsTransformer '
                            'to turn the raster data into uint8 data'.format(
                                rv.TF_DEEPLAB))

        tf_examples = make_tf_examples(data, self.class_map)

        base_uri = self.backend_config.training_data_uri
        split = '{}-{}'.format(scene.id, uuid.uuid4())
        record_path = join(base_uri, '{}.record'.format(split))
        record_path = get_local_path(record_path, tmp_dir)

        make_dir(record_path, use_dirname=True)
        write_tf_record(tf_examples, record_path)

        return record_path

    def process_sceneset_results(self, training_results: List[str],
                                 validation_results: List[str],
                                 tmp_dir: str) -> None:
        """Merge TFRecord files from individual scenes into two at-large files
        (one for training data and one for validation data).

        Args:
             training_results: A list of paths to TFRecords containing
                  training data.
             validation_results: A list of paths to TFRecords
                  containing validation data.
             tmp_dir: (str) temporary directory to use
        Returns:
             None

        """
        base_uri = self.backend_config.training_data_uri
        chip_suffix = str(uuid.uuid4()).split('-')[0]
        training_record_path = get_record_uri(base_uri, TRAIN, chip_suffix)
        training_record_path_local = get_local_path(training_record_path,
                                                    tmp_dir)
        validation_record_path = get_record_uri(base_uri, VALIDATION,
                                                chip_suffix)
        validation_record_path_local = get_local_path(validation_record_path,
                                                      tmp_dir)

        make_dir(training_record_path_local, use_dirname=True)
        make_dir(validation_record_path_local, use_dirname=True)  # sic
        merge_tf_records(training_record_path_local, training_results)
        merge_tf_records(validation_record_path_local, validation_results)
        upload_or_copy(training_record_path_local, training_record_path)
        upload_or_copy(validation_record_path_local, validation_record_path)

        if self.backend_config.debug:

            def _make_debug_chips(split, record_path_local):
                zip_path = join(base_uri, '{}-debug'.format(split),
                                chip_suffix)
                zip_path_local = get_local_path(zip_path, tmp_dir)

                debug_dir = join(tmp_dir, '{}-debug'.format(split),
                                 chip_suffix)
                make_debug_images(
                    record_path_local, debug_dir, self.class_map,
                    self.task_config.chip_options.debug_chip_probability)
                shutil.make_archive(zip_path_local, 'zip', debug_dir)

                upload_or_copy('{}.zip'.format(zip_path_local),
                               '{}.zip'.format(zip_path))

            if training_results:
                _make_debug_chips(TRAIN, training_record_path_local)
            if validation_results:
                _make_debug_chips(VALIDATION, validation_record_path_local)

    def train(self, tmp_dir: str) -> None:
        """Train a DeepLab model the task and backend config.

        Args:
            tmp_dir: (str) temporary directory to use

        Returns:
             None
        """
        train_py = self.backend_config.script_locations.train_py
        eval_py = self.backend_config.script_locations.eval_py
        export_py = self.backend_config.script_locations.export_py

        # Setup local input and output directories
        log.info('Setting up local input and output directories')
        train_logdir = self.backend_config.training_output_uri
        train_logdir_local = get_local_path(train_logdir, tmp_dir)
        dataset_dir = get_record_dir(self.backend_config.training_data_uri,
                                     TRAIN)
        dataset_dir_local = get_local_path(dataset_dir, tmp_dir)
        make_dir(tmp_dir)
        make_dir(train_logdir_local)
        make_dir(dataset_dir_local)

        # Download training data
        log.info('Downloading training data')
        for i, record_file in enumerate(list_paths(dataset_dir)):
            download_if_needed(record_file, tmp_dir)

        # Download and untar initial checkpoint.
        log.info('Downloading and untarring initial checkpoint')
        tf_initial_checkpoints_uri = self.backend_config.pretrained_model_uri
        download_if_needed(tf_initial_checkpoints_uri, tmp_dir)
        tfic_tarball = get_local_path(tf_initial_checkpoints_uri, tmp_dir)
        tfic_dir = os.path.dirname(tfic_tarball)
        with tarfile.open(tfic_tarball, 'r:gz') as tar:
            tar.extractall(tfic_dir)
        tfic_ckpt = glob.glob('{}/*/*.index'.format(tfic_dir))[0]
        tfic_ckpt = tfic_ckpt[0:-len('.index')]

        # Restart support
        train_restart_dir = self.backend_config.train_options.train_restart_dir
        if type(train_restart_dir) is not str or len(train_restart_dir) == 0:
            train_restart_dir = train_logdir

        # Get output from potential previous run so we can resume training.
        if type(train_restart_dir) is str and len(
                train_restart_dir
        ) > 0 and not self.backend_config.train_options.replace_model:
            sync_from_dir(train_restart_dir, train_logdir_local)

            # Need to update model_checkpoint_path in the checkpoint file,
            # since it has the absolute paths from the previous run which
            # was using a different temporary directory on another machine.
            # If Deeplab could save relative paths instead (like the Object
            # Detection API does), then we wouldn't need to do this.
            checkpoint_path = join(train_logdir_local, 'checkpoint')
            if file_exists(checkpoint_path):
                latest_checkpoint = get_latest_checkpoint(train_logdir_local)
                with open(checkpoint_path, 'w') as cf:
                    cf.write('model_checkpoint_path: \"{}\"'.format(
                        latest_checkpoint))
        else:
            if self.backend_config.train_options.replace_model:
                if os.path.exists(train_logdir_local):
                    shutil.rmtree(train_logdir_local)
                make_dir(train_logdir_local)

        # Periodically synchronize with remote
        sync = start_sync(
            train_logdir_local,
            train_logdir,
            sync_interval=self.backend_config.train_options.sync_interval)

        with sync:
            # Setup TFDL config
            tfdl_config = json_format.ParseDict(
                self.backend_config.tfdl_config, TrainingParametersMsg())
            log.info('tfdl_config={}'.format(tfdl_config))
            log.info('Training steps={}'.format(
                tfdl_config.training_number_of_steps))

            # Additional training options
            max_class = max(
                list(map(lambda c: c.id, self.class_map.get_items())))
            num_classes = len(self.class_map.get_items())
            num_classes = max(max_class, num_classes) + 1
            (train_args, train_env) = get_training_args(
                train_py, train_logdir_local, tfic_ckpt, dataset_dir_local,
                num_classes, tfdl_config)

            # Start training
            log.info('Starting training process')
            log.info(' '.join(train_args))
            train_process = Popen(train_args, env=train_env)
            terminate_at_exit(train_process)

            if self.backend_config.train_options.do_monitoring:
                # Start tensorboard
                log.info('Starting tensorboard process')
                tensorboard_process = Popen(
                    ['tensorboard', '--logdir={}'.format(train_logdir_local)])
                terminate_at_exit(tensorboard_process)

            if self.backend_config.train_options.do_eval:
                # Start eval script
                log.info('Starting eval script')
                eval_logdir = train_logdir_local
                eval_args = get_evaluation_args(eval_py, train_logdir_local,
                                                dataset_dir_local, eval_logdir,
                                                tfdl_config)
                eval_process = Popen(eval_args, env=train_env)
                terminate_at_exit(eval_process)

            # Wait for training and tensorboard
            log.info('Waiting for training and tensorboard processes')
            train_process.wait()
            if self.backend_config.train_options.do_monitoring:
                tensorboard_process.terminate()

            # Export frozen graph
            log.info(
                'Exporting frozen graph ({}/model)'.format(train_logdir_local))
            export_args = get_export_args(export_py, train_logdir_local,
                                          num_classes, tfdl_config)
            export_process = Popen(export_args)
            terminate_at_exit(export_process)
            export_process.wait()

            # Package up the model files for usage as fine tuning checkpoints
            fine_tune_checkpoint_name = self.backend_config.fine_tune_checkpoint_name
            latest_checkpoints = get_latest_checkpoint(train_logdir_local)
            model_checkpoint_files = glob.glob(
                '{}*'.format(latest_checkpoints))
            inference_graph_path = os.path.join(train_logdir_local, 'model')

            with RVConfig.get_tmp_dir() as tmp_dir:
                model_dir = os.path.join(tmp_dir, fine_tune_checkpoint_name)
                make_dir(model_dir)
                model_tar = os.path.join(
                    train_logdir_local,
                    '{}.tar.gz'.format(fine_tune_checkpoint_name))
                shutil.copy(inference_graph_path,
                            '{}/frozen_inference_graph.pb'.format(model_dir))
                for path in model_checkpoint_files:
                    shutil.copy(path, model_dir)
                with tarfile.open(model_tar, 'w:gz') as tar:
                    tar.add(model_dir, arcname=os.path.basename(model_dir))

        # Perform final sync
        sync_to_dir(train_logdir_local, train_logdir, delete=False)

    def load_model(self, tmp_dir: str):
        """Load the model in preparation for one or more prediction calls.

        Args:
             tmp_dir: (str) temporary directory to use
        """
        # noqa Courtesy of https://github.com/tensorflow/models/blob/cbbb2ffcde66e646d4a47628ffe2ece2322b64e8/research/deeplab/deeplab_demo.ipynb
        import tensorflow as tf
        if self.sess is None:
            FROZEN_GRAPH_NAME = download_if_needed(
                self.backend_config.model_uri, tmp_dir)
            graph = tf.Graph()
            with open(FROZEN_GRAPH_NAME, 'rb') as data:
                graph_def = tf.GraphDef.FromString(data.read())
            with graph.as_default():
                tf.import_graph_def(graph_def, name='')
            self.sess = tf.Session(graph=graph)

    def predict(self, chips: np.ndarray, windows: List[Box],
                tmp_dir: str) -> SemanticSegmentationLabels:
        """Predict using an already-trained DeepLab model.

        Args:
            chips: An np.ndarray containing the image data in a batch of size 1.
            tmp_dir: (str) temporary directory to use

        Returns:
             SemanticSegmentationLabels object with predictions for a single chip
        """
        self.load_model(tmp_dir)
        label_arr = self.sess.run(
            OUTPUT_TENSOR_NAME, feed_dict={INPUT_TENSOR_NAME: [chips[0]]})[0]

        # Return "trivial" instance of SemanticSegmentationLabels that holds a single
        # window and has ability to get labels for that one window.
        def label_fn(_window):
            if _window == windows[0]:
                return label_arr
            else:
                raise ValueError('Trying to get labels for unknown window.')

        labels = SemanticSegmentationLabels(windows, label_fn)

        return labels



# flake8: noqa

from rastervision.data.label_store.label_store import *
from rastervision.data.label_store.label_store_config import *
from rastervision.data.label_store.chip_classification_geojson_store import *
from rastervision.data.label_store.chip_classification_geojson_store_config import *
from rastervision.data.label_store.object_detection_geojson_store import *
from rastervision.data.label_store.object_detection_geojson_store_config import *
from rastervision.data.label_store.semantic_segmentation_raster_store import *
from rastervision.data.label_store.semantic_segmentation_raster_store_config import *



import os
from copy import deepcopy

import rastervision as rv
from rastervision.data.label_store import (
    LabelStoreConfig, LabelStoreConfigBuilder, ObjectDetectionGeoJSONStore)


class ObjectDetectionGeoJSONStoreConfig(LabelStoreConfig):
    def __init__(self, uri=None):
        super().__init__(store_type=rv.OBJECT_DETECTION_GEOJSON)
        self.uri = uri

    def to_proto(self):
        msg = super().to_proto()
        if self.uri:
            msg.uri = self.uri
        return msg

    def for_prediction(self, label_uri):
        return self.to_builder() \
                   .with_uri(label_uri) \
                   .build()

    def create_store(self, task_config, extent, crs_transformer, tmp_dir):
        return ObjectDetectionGeoJSONStore(self.uri, crs_transformer,
                                           task_config.class_map)

    def update_for_command(self, command_type, experiment_config,
                           context=None):
        if command_type == rv.PREDICT:
            if not self.uri:
                # Construct the  URI for this prediction store,
                # using the scene ID.
                root = experiment_config.predict_uri
                uri = None
                for c in context:
                    if isinstance(c, rv.SceneConfig):
                        uri = os.path.join(root, '{}.json'.format(c.id))
                if uri:
                    self.uri = uri
                else:
                    raise rv.ConfigError(
                        'ObjectDetectionGeoJSONStoreConfig has no '
                        'URI set, and is not associated with a SceneConfig.')

    def report_io(self, command_type, io_def):
        if command_type == rv.PREDICT:
            io_def.add_output(self.uri)

        if command_type == rv.EVAL:
            if self.uri:
                io_def.add_input(self.uri)
            else:
                msg = 'No URI set for ObjectDetectionGeoJSONStoreConfig'
                io_def.add_missing(msg)


class ObjectDetectionGeoJSONStoreConfigBuilder(LabelStoreConfigBuilder):
    def __init__(self, prev=None):
        config = {}
        if prev:
            config = {'uri': prev.uri}

        super().__init__(ObjectDetectionGeoJSONStoreConfig, config)

    def from_proto(self, msg):
        return self \
            .with_uri(msg.uri)

    def with_uri(self, uri):
        """Set URI for a GeoJSON used to read/write predictions."""
        b = deepcopy(self)
        b.config['uri'] = uri
        return b



from os.path import (join, isfile, basename, dirname)
import uuid
import zipfile
import glob
import logging
import json
from subprocess import Popen
import os
import csv
import time
import datetime

import matplotlib
matplotlib.use('Agg')  # noqa
import matplotlib.pyplot as plt
import torch
from torch import optim
from torch.utils.tensorboard import SummaryWriter
from torch.optim.lr_scheduler import CyclicLR
import numpy as np

from rastervision.utils.files import (get_local_path, make_dir, upload_or_copy,
                                      list_paths, download_if_needed,
                                      sync_from_dir, sync_to_dir, str_to_file,
                                      zipdir, file_to_json, json_to_file)
from rastervision.utils.misc import save_img
from rastervision.backend import Backend
from rastervision.data.label import SemanticSegmentationLabels
from rastervision.utils.misc import terminate_at_exit
from rastervision.backend.torch_utils.semantic_segmentation.plot import plot_xy
from rastervision.backend.torch_utils.semantic_segmentation.data import build_databunch
from rastervision.backend.torch_utils.semantic_segmentation.train import (
    train_epoch, validate_epoch)
from rastervision.backend.torch_utils.semantic_segmentation.model import (
    get_model)

log = logging.getLogger(__name__)


def make_debug_chips(databunch, class_map, tmp_dir, train_uri, max_count=30):
    """Save debug chips for a Databunch for a semantic segmentation dataset.

    This saves a plot for each example in the training and validation sets into
    train-debug-chips.zip and valid-debug-chips.zip under the train_uri. This
    is useful for making sure we are feeding correct data into the model.

    Args:
        databunch: DataBunch for semantic segmentation
        class_map: (rv.ClassMap) class map used to map class ids to colors
        tmp_dir: (str) path to temp directory
        train_uri: (str) URI of root of training output
        max_count: (int) maximum number of chips to generate. If None,
            generates all of them.
    """

    def _make_debug_chips(split):
        debug_chips_dir = join(tmp_dir, '{}-debug-chips'.format(split))
        zip_path = join(tmp_dir, '{}-debug-chips.zip'.format(split))
        zip_uri = join(train_uri, '{}-debug-chips.zip'.format(split))
        make_dir(debug_chips_dir)
        ds = databunch.train_ds if split == 'train' else databunch.valid_ds
        for i, (x, y) in enumerate(ds):
            if i >= max_count:
                break

            fig, ax = plt.subplots(1)
            plot_xy(ax, x, class_map, y=y)
            plt.savefig(
                join(debug_chips_dir, '{}.png'.format(i)), figsize=(6, 6))
            plt.close()

        zipdir(debug_chips_dir, zip_path)
        upload_or_copy(zip_path, zip_uri)

    _make_debug_chips('train')
    _make_debug_chips('valid')


class PyTorchSemanticSegmentation(Backend):
    """Semantic segmentation backend using PyTorch and fastai."""

    def __init__(self, task_config, backend_opts, train_opts):
        """Constructor.

        Args:
            task_config: (SemanticSegmentationConfig)
            backend_opts: (simple_backend_config.BackendOptions)
            train_opts: (pytorch_semantic_segmentation_backend_config.TrainOptions)
        """
        self.task_config = task_config
        self.backend_opts = backend_opts
        self.train_opts = train_opts
        self.inf_learner = None

        torch_cache_dir = '/opt/data/torch-cache'
        os.environ['TORCH_HOME'] = torch_cache_dir

        self.model = None
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        log.info('Device = {}'.format(self.device))
        # TODO move this into the SemanticSegmentation RV task
        self.class_map = self.task_config.class_map.copy()
        self.class_map.add_nodata_item()

    def log_options(self):
        log.info('backend_opts:\n' +
                 json.dumps(self.backend_opts.__dict__, indent=4))
        log.info('train_opts:\n' +
                 json.dumps(self.train_opts.__dict__, indent=4))

    def process_scene_data(self, scene, data, tmp_dir):
        """Make training chips for a scene.

        This writes a set of image chips to {scene_id}/img/{scene_id}-{ind}.png
        and corresponding label chips to {scene_id}/labels/{scene_id}-{ind}.png.

        Args:
            scene: (rv.data.Scene)
            data: (rv.data.Dataset)
            tmp_dir: (str) path to temp directory

        Returns:
            (str) path to directory with scene chips {tmp_dir}/{scene_id}
        """
        scene_dir = join(tmp_dir, str(scene.id))
        img_dir = join(scene_dir, 'img')
        labels_dir = join(scene_dir, 'labels')

        make_dir(img_dir)
        make_dir(labels_dir)

        for ind, (chip, window, labels) in enumerate(data):
            chip_path = join(img_dir, '{}-{}.png'.format(scene.id, ind))
            label_path = join(labels_dir, '{}-{}.png'.format(scene.id, ind))

            label_im = labels.get_label_arr(window).astype(np.uint8)
            save_img(label_im, label_path)
            save_img(chip, chip_path)

        return scene_dir

    def process_sceneset_results(self, training_results, validation_results,
                                 tmp_dir):
        """Write zip file with chips for a set of scenes.

        This writes a zip file for a group of scenes at {chip_uri}/{uuid}.zip containing:
        train/img/{scene_id}-{ind}.png
        train/labels/{scene_id}-{ind}.png
        val/img/{scene_id}-{ind}.png
        val/labels/{scene_id}-{ind}.png

        This method is called once per instance of the chip command.
        A number of instances of the chip command can run simultaneously to
        process chips in parallel. The uuid in the path above is what allows
        separate instances to avoid overwriting each others' output.

        Args:
            training_results: list of directories generated by process_scene_data
                that all hold training chips
            validation_results: list of directories generated by process_scene_data
                that all hold validation chips
        """
        self.log_options()

        group = str(uuid.uuid4())
        group_uri = join(self.backend_opts.chip_uri, '{}.zip'.format(group))
        group_path = get_local_path(group_uri, tmp_dir)
        make_dir(group_path, use_dirname=True)

        with zipfile.ZipFile(group_path, 'w', zipfile.ZIP_DEFLATED) as zipf:

            def _write_zip(results, split):
                for scene_dir in results:
                    scene_paths = glob.glob(join(scene_dir, '**/*.png'))
                    for p in scene_paths:
                        zipf.write(
                            p,
                            join(
                                '{}/{}'.format(split,
                                               dirname(p).split('/')[-1]),
                                basename(p)))

            _write_zip(training_results, 'train')
            _write_zip(validation_results, 'valid')

        upload_or_copy(group_path, group_uri)

    def train(self, tmp_dir):
        """Train a model.

        This downloads any previous output saved to the train_uri,
        starts training (or resumes from a checkpoint), periodically
        syncs contents of train_dir to train_uri and after training finishes.

        Args:
            tmp_dir: (str) path to temp directory
        """
        self.log_options()

        # Sync output of previous training run from cloud.
        train_uri = self.backend_opts.train_uri
        train_dir = get_local_path(train_uri, tmp_dir)
        make_dir(train_dir)
        sync_from_dir(train_uri, train_dir)

        # Get zip file for each group, and unzip them into chip_dir.
        chip_dir = join(tmp_dir, 'chips')
        make_dir(chip_dir)
        for zip_uri in list_paths(self.backend_opts.chip_uri, 'zip'):
            zip_path = download_if_needed(zip_uri, tmp_dir)
            with zipfile.ZipFile(zip_path, 'r') as zipf:
                zipf.extractall(chip_dir)

        # Setup data loader.
        batch_size = self.train_opts.batch_size
        chip_size = self.task_config.chip_size
        class_names = self.class_map.get_class_names()
        databunch = build_databunch(chip_dir, chip_size, batch_size,
                                    class_names)
        log.info(databunch)
        num_labels = len(databunch.label_names)
        if self.train_opts.debug:
            make_debug_chips(databunch, self.class_map, tmp_dir, train_uri)

        # Setup model
        num_labels = len(databunch.label_names)
        model = get_model(
            self.train_opts.model_arch, num_labels, pretrained=True)
        model = model.to(self.device)
        model_path = join(train_dir, 'model')

        # Load weights from a pretrained model.
        pretrained_uri = self.backend_opts.pretrained_uri
        if pretrained_uri:
            log.info('Loading weights from pretrained_uri: {}'.format(
                pretrained_uri))
            pretrained_path = download_if_needed(pretrained_uri, tmp_dir)
            model.load_state_dict(
                torch.load(pretrained_path, map_location=self.device))

        # Possibly resume training from checkpoint.
        start_epoch = 0
        train_state_path = join(train_dir, 'train_state.json')
        if isfile(train_state_path):
            log.info('Resuming from checkpoint: {}\n'.format(model_path))
            train_state = file_to_json(train_state_path)
            start_epoch = train_state['epoch'] + 1
            model.load_state_dict(
                torch.load(model_path, map_location=self.device))

        # Write header of log CSV file.
        metric_names = ['precision', 'recall', 'f1']
        log_path = join(train_dir, 'log.csv')
        if not isfile(log_path):
            with open(log_path, 'w') as log_file:
                log_writer = csv.writer(log_file)
                row = ['epoch', 'time', 'train_loss'] + metric_names
                log_writer.writerow(row)

        # Setup Tensorboard logging.
        if self.train_opts.log_tensorboard:
            log_dir = join(train_dir, 'tb-logs')
            make_dir(log_dir)
            tb_writer = SummaryWriter(log_dir=log_dir)
            if self.train_opts.run_tensorboard:
                log.info('Starting tensorboard process')
                tensorboard_process = Popen(
                    ['tensorboard', '--logdir={}'.format(log_dir)])
                terminate_at_exit(tensorboard_process)

        # Setup optimizer, loss, and LR scheduler.
        loss_fn = torch.nn.CrossEntropyLoss()
        lr = self.train_opts.lr
        opt = optim.Adam(model.parameters(), lr=lr)
        step_scheduler, epoch_scheduler = None, None
        num_epochs = self.train_opts.num_epochs

        if self.train_opts.one_cycle and num_epochs > 1:
            steps_per_epoch = len(databunch.train_ds) // batch_size
            total_steps = num_epochs * steps_per_epoch
            step_size_up = (num_epochs // 2) * steps_per_epoch
            step_size_down = total_steps - step_size_up
            step_scheduler = CyclicLR(
                opt,
                base_lr=lr / 10,
                max_lr=lr,
                step_size_up=step_size_up,
                step_size_down=step_size_down,
                cycle_momentum=False)
            for _ in range(start_epoch * steps_per_epoch):
                step_scheduler.step()

        # Training loop.
        for epoch in range(start_epoch, num_epochs):
            # Train one epoch.
            log.info('-----------------------------------------------------')
            log.info('epoch: {}'.format(epoch))
            start = time.time()
            train_loss = train_epoch(model, self.device, databunch.train_dl,
                                     opt, loss_fn, step_scheduler)
            if epoch_scheduler:
                epoch_scheduler.step()
            log.info('train loss: {}'.format(train_loss))

            # Validate one epoch.
            metrics = validate_epoch(model, self.device, databunch.valid_dl,
                                     num_labels)
            log.info('validation metrics: {}'.format(metrics))

            # Print elapsed time for epoch.
            end = time.time()
            epoch_time = datetime.timedelta(seconds=end - start)
            log.info('epoch elapsed time: {}'.format(epoch_time))

            # Save model and state.
            torch.save(model.state_dict(), model_path)
            train_state = {'epoch': epoch}
            json_to_file(train_state, train_state_path)

            # Append to log CSV file.
            with open(log_path, 'a') as log_file:
                log_writer = csv.writer(log_file)
                row = [epoch, epoch_time, train_loss]
                row += [metrics[k] for k in metric_names]
                log_writer.writerow(row)

            # Write to Tensorboard log.
            if self.train_opts.log_tensorboard:
                for key, val in metrics.items():
                    tb_writer.add_scalar(key, val, epoch)
                tb_writer.add_scalar('train_loss', train_loss, epoch)
                for name, param in model.named_parameters():
                    tb_writer.add_histogram(name, param, epoch)

            if (train_uri.startswith('s3://')
                    and (((epoch + 1) % self.train_opts.sync_interval) == 0)):
                sync_to_dir(train_dir, train_uri)

        # Close Tensorboard.
        if self.train_opts.log_tensorboard:
            tb_writer.close()
            if self.train_opts.run_tensorboard:
                tensorboard_process.terminate()

        # Since model is exported every epoch, we need some other way to
        # show that training is finished.
        str_to_file('done!', self.backend_opts.train_done_uri)

        # Sync output to cloud.
        sync_to_dir(train_dir, self.backend_opts.train_uri)

    def load_model(self, tmp_dir):
        """Load the model in preparation for one or more prediction calls."""
        if self.model is None:
            model_uri = self.backend_opts.model_uri
            model_path = download_if_needed(model_uri, tmp_dir)

            num_classes = len(self.class_map)
            model = get_model(
                self.train_opts.model_arch, num_classes, pretrained=True)
            model = model.to(self.device)
            model.load_state_dict(
                torch.load(model_path, map_location=self.device))
            self.model = model

    def predict(self, chips, windows, tmp_dir):
        """Return a prediction for a single chip.

        Args:
            chips: (numpy.ndarray) of shape (1, height, width, nb_channels)
                containing a single imagery chip
            windows: List containing a single (Box) window which is aligned
                with the chip

        Return:
            (SemanticSegmentationLabels) containing predictions
        """
        self.load_model(tmp_dir)

        chips = torch.Tensor(chips).permute((0, 3, 1, 2)) / 255.
        chips = chips.to(self.device)
        model = self.model.eval()

        with torch.no_grad():
            out = model(chips)['out'].cpu()

        def label_fn(_window):
            if _window == windows[0]:
                return out[0].argmax(0).squeeze().numpy()
            else:
                raise ValueError('Trying to get labels for unknown window.')

        return SemanticSegmentationLabels(windows, label_fn)



import rastervision as rv
from rastervision.evaluation import ObjectDetectionEvaluator
from rastervision.evaluation \
    import (ClassificationEvaluatorConfig, ClassificationEvaluatorConfigBuilder)


class ObjectDetectionEvaluatorConfig(ClassificationEvaluatorConfig):
    def __init__(self, class_map, output_uri=None):
        super().__init__(rv.OBJECT_DETECTION_EVALUATOR, class_map, output_uri)

    def create_evaluator(self):
        return ObjectDetectionEvaluator(self.class_map, self.output_uri)


class ObjectDetectionEvaluatorConfigBuilder(
        ClassificationEvaluatorConfigBuilder):
    def __init__(self, prev=None):
        super().__init__(ObjectDetectionEvaluatorConfig, prev)



import random


class TrainingData(object):
    """A set of chips, windows, and labels used to train a model."""

    def __init__(self):
        """Construct a new TrainingData."""
        self.chips = []
        self.windows = []
        self.labels = []

    def append(self, chip, window, labels):
        """Append a chip and associated labels to the dataset.

        Args:
            chip: [height, width, channels] numpy array
            window: Box with coordinates of chip
            labels: Labels
        """
        self.chips.append(chip)
        self.windows.append(window)
        self.labels.append(labels)

    def __iter__(self):
        return zip(self.chips, self.windows, self.labels)

    def shuffle(self):
        """Randomly shuffle the chips and labels in-place.

        This maintains the correspondence between chips and labels.
        """
        if len(self.chips) > 0:
            chip_windows_labels = list(self)
            random.shuffle(chip_windows_labels)
            # Unzip the list.
            self.chips, self.windows, self.labels = zip(*chip_windows_labels)



from abc import abstractmethod

import rastervision as rv
from rastervision.core.config import (Config, ConfigBuilder,
                                      BundledConfigMixin)


class AnalyzerConfig(BundledConfigMixin, Config):
    def __init__(self, analyzer_type):
        self.analyzer_type = analyzer_type

    @abstractmethod
    def create_analyzer(self):
        """Create the Transformer that this configuration represents
        """
        pass

    def to_builder(self):
        return rv._registry.get_config_builder(rv.ANALYZER,
                                               self.analyzer_type)(self)

    @staticmethod
    def builder(analyzer_type):
        return rv._registry.get_config_builder(rv.ANALYZER, analyzer_type)()

    @staticmethod
    def from_proto(msg):
        """Creates a TaskConfig from the specificed protobuf message
        """
        return rv._registry.get_config_builder(rv.ANALYZER, msg.analyzer_type)() \
                           .from_proto(msg) \
                           .build()


class AnalyzerConfigBuilder(ConfigBuilder):
    pass



from os.path import join
from copy import deepcopy
from abc import abstractmethod
import json

from google.protobuf import struct_pb2

import rastervision as rv
from rastervision.backend import (BackendConfig, BackendConfigBuilder)
from rastervision.protos.backend_pb2 import BackendConfig as BackendConfigMsg


class BackendOptions():
    """Options that pertain to backends created using SimpleBackendConfig."""

    def __init__(self,
                 chip_uri=None,
                 train_uri=None,
                 train_done_uri=None,
                 model_uri=None,
                 pretrained_uri=None):
        self.chip_uri = chip_uri
        self.train_uri = train_uri
        self.train_done_uri = train_done_uri
        self.model_uri = model_uri
        self.pretrained_uri = pretrained_uri


class SimpleBackendConfig(BackendConfig):
    """A simplified BackendConfig interface.

    This class can be subclassed to created BackendConfigs with less effort
    and a small loss of flexibility when compared to directly subclassing
    BackendConfig. See subclasses of this for examples of how to write your
    own subclass.
    """

    def __init__(self, backend_opts, train_opts):
        """Constructor.

        Args:
            backend_opts: (BackendOptions)
            train_opts: (train_opts_class) object containing options that are
                set by with_train_options()
        """
        super().__init__(self.backend_type)
        self.backend_opts = backend_opts
        self.train_opts = train_opts

    @property
    @abstractmethod
    def train_opts_class(self):
        """The class that holds options set by with_train_options in the builder."""
        pass

    @property
    @abstractmethod
    def backend_type(self):
        """The string representing this backend used in the registry."""
        pass

    @property
    @abstractmethod
    def backend_class(self):
        """The class of the actual Backend that this class configures."""
        pass

    def to_proto(self):
        config = {}
        for k, v in self.backend_opts.__dict__.items():
            config[k] = v
        for k, v in self.train_opts.__dict__.items():
            config[k] = v

        custom_config = struct_pb2.Struct()
        custom_config['json'] = json.dumps(config)
        msg = BackendConfigMsg(
            backend_type=self.backend_type, custom_config=custom_config)
        return msg

    def create_backend(self, task_config):
        return self.backend_class(task_config, self.backend_opts,
                                  self.train_opts)

    def update_for_command(self, command_type, experiment_config,
                           context=None):
        super().update_for_command(command_type, experiment_config, context)

        if command_type == rv.CHIP:
            self.backend_opts.chip_uri = join(experiment_config.chip_uri,
                                              'chips')
        elif command_type == rv.TRAIN:
            self.backend_opts.train_uri = experiment_config.train_uri
            self.backend_opts.model_uri = join(experiment_config.train_uri,
                                               'model')
            self.backend_opts.train_done_uri = join(
                experiment_config.train_uri, 'done.txt')

    def report_io(self, command_type, io_def):
        super().report_io(command_type, io_def)

        if command_type == rv.CHIP:
            io_def.add_output(self.backend_opts.chip_uri)
        elif command_type == rv.TRAIN:
            io_def.add_input(self.backend_opts.chip_uri)
            io_def.add_output(self.backend_opts.model_uri)
            io_def.add_output(self.backend_opts.train_done_uri)
        elif command_type in [rv.PREDICT, rv.BUNDLE]:
            if not self.backend_opts.model_uri:
                io_def.add_missing('Missing model_uri.')
            else:
                io_def.add_input(self.backend_opts.model_uri)
                io_def.add_input(self.backend_opts.train_done_uri)
        elif command_type == rv.EVAL:
            io_def.add_input()

    def save_bundle_files(self, bundle_dir):
        model_uri = self.backend_opts.model_uri
        if not model_uri:
            raise rv.ConfigError('model_uri is not set.')
        local_path, base_name = self.bundle_file(model_uri, bundle_dir)
        new_config = self.to_builder() \
                         .with_model_uri(base_name) \
                         .build()
        return (new_config, [local_path])

    def load_bundle_files(self, bundle_dir):
        model_uri = self.backend_opts.model_uri
        if not model_uri:
            raise rv.ConfigError('model_uri is not set.')
        local_model_uri = join(bundle_dir, model_uri)
        return self.to_builder() \
                   .with_model_uri(local_model_uri) \
                   .build()


class SimpleBackendConfigBuilder(BackendConfigBuilder):
    """A simplified BackendConfigBuilder interface.

    This class can be subclassed to created BackendConfigBuilders with less
    effort and a small loss of flexibility when compared to directly
    subclassing BackendConfigBuilder. See subclasses of this for examples of
    how to write your own subclass.
    """

    def __init__(self, prev_config=None):
        self.backend_opts = BackendOptions()
        self.train_opts = self.config_class.train_opts_class()

        if prev_config:
            self.backend_opts = prev_config.backend_opts
            self.train_opts = prev_config.train_opts

        super().__init__(self.config_class.backend_type, self.config_class)
        self.require_task = prev_config is None

    @property
    @abstractmethod
    def config_class(self):
        """The corresponding Config class for this builder."""
        pass

    @abstractmethod
    def with_train_options(self):
        """Sets the training options which are passed to the Backend."""
        pass

    def build(self):
        self.validate()
        return self.config_class(self.backend_opts, self.train_opts)

    def from_proto(self, msg):
        b = super().from_proto(msg)
        custom_config = msg.custom_config

        if 'json' in custom_config:
            config = json.loads(custom_config['json'])
        else:
            config = custom_config

        for k in self.backend_opts.__dict__.keys():
            if k in config:
                setattr(b.backend_opts, k, config[k])
        for k in self.train_opts.__dict__.keys():
            if k in config:
                setattr(b.train_opts, k, config[k])
        b.require_task = None
        return b

    def validate(self):
        super().validate()

        if self.require_task and not self.task:
            raise rv.ConfigError('You must specify the task this backend '
                                 "is for - use 'with_task'.")

        return True

    def _process_task(self):
        return self

    def with_model_uri(self, model_uri):
        b = deepcopy(self)
        b.backend_opts.model_uri = model_uri
        return b

    def with_pretrained_uri(self, pretrained_uri):
        b = deepcopy(self)
        b.backend_opts.pretrained_uri = pretrained_uri
        return b



class DataBunch():
    def __init__(self, train_ds, train_dl, valid_ds, valid_dl, label_names):
        self.train_ds = train_ds
        self.train_dl = train_dl
        self.valid_ds = valid_ds
        self.valid_dl = valid_dl
        self.label_names = label_names

    def __repr__(self):
        rep = ''
        rep += 'train_ds: {} items\n'.format(len(self.train_ds))
        rep += 'valid_ds: {} items\n'.format(len(self.valid_ds))
        rep += 'label_names: ' + ','.join(self.label_names)
        return rep



import os
import shutil
import urllib
import urllib.request
from datetime import datetime

from rastervision.filesystem import (FileSystem, NotReadableError,
                                     NotWritableError)
from urllib.parse import urlparse


class HttpFileSystem(FileSystem):
    @staticmethod
    def matches_uri(uri: str, mode: str) -> bool:
        parsed_uri = urlparse(uri)
        return parsed_uri.scheme in ['http', 'https']

    @staticmethod
    def file_exists(uri: str, include_dir: bool = True) -> bool:
        try:
            response = urllib.request.urlopen(uri)
            if response.getcode() == 200:
                return int(response.headers['content-length']) > 0
            else:
                return False  # pragma: no cover
        except urllib.error.URLError:
            return False

    @staticmethod
    def read_str(uri: str) -> str:
        return HttpFileSystem.read_bytes(uri).decode('utf8')

    @staticmethod
    def read_bytes(uri: str) -> bytes:
        with urllib.request.urlopen(uri) as req:
            return req.read()

    @staticmethod
    def write_str(uri: str, data: str) -> None:
        raise NotWritableError('Could not write {}'.format(uri))

    @staticmethod
    def write_bytes(uri: str, data: bytes) -> None:
        raise NotWritableError('Could not write {}'.format(uri))

    @staticmethod
    def sync_to_dir(src_dir_uri: str, dest_dir_uri: str,
                    delete: bool = False) -> None:
        raise NotWritableError('Could not write {}'.format(dest_dir_uri))

    @staticmethod
    def sync_from_dir(src_dir_uri: str,
                      dest_dir_uri: str,
                      delete: bool = False) -> None:
        raise NotReadableError(
            'Cannot read directory from HTTP {}'.format(src_dir_uri))

    @staticmethod
    def copy_to(src_path: str, dst_uri: str) -> None:
        raise NotWritableError('Could not write {}'.format(dst_uri))

    @staticmethod
    def copy_from(uri: str, path: str) -> None:
        with urllib.request.urlopen(uri) as response:
            with open(path, 'wb') as out_file:
                try:
                    shutil.copyfileobj(response, out_file)
                except Exception:  # pragma: no cover
                    raise NotReadableError('Could not read {}'.format(uri))

    @staticmethod
    def local_path(uri: str, download_dir: str) -> None:
        parsed_uri = urlparse(uri)
        path = os.path.join(download_dir, 'http', parsed_uri.netloc,
                            parsed_uri.path[1:])
        return path

    @staticmethod
    def last_modified(uri: str) -> datetime:
        return None

    @staticmethod
    def list_paths(uri, suffix=None):  # pragma: no cover
        raise NotImplementedError()



from abc import abstractmethod

import rastervision as rv
from rastervision.core import (Config, ConfigBuilder)


class EvaluatorConfig(Config):
    def __init__(self, evaluator_type):
        self.evaluator_type = evaluator_type

    @abstractmethod
    def create_evaluator(self):
        """Create the Evaluator that this configuration represents"""
        pass

    def to_builder(self):
        return rv._registry.get_config_builder(rv.EVALUATOR,
                                               self.evaluator_type)(self)

    @staticmethod
    def builder(evaluator_type):
        return rv._registry.get_config_builder(rv.EVALUATOR, evaluator_type)()

    @staticmethod
    def from_proto(msg):
        """Creates a EvaluatorConfig from the specificed protobuf message
        """
        return rv._registry.get_config_builder(rv.EVALUATOR, msg.evaluator_type)() \
                           .from_proto(msg) \
                           .build()


class EvaluatorConfigBuilder(ConfigBuilder):
    pass



import copy

from rastervision.data.vector_source.label_maker.filter import create_filter


class ClassInferenceOptions():
    def __init__(self,
                 class_map=None,
                 class_id_to_filter=None,
                 default_class_id=1):
        self.class_map = class_map
        self.class_id_to_filter = class_id_to_filter
        self.default_class_id = default_class_id


class ClassInference():
    """Infers missing class_ids from GeoJSON features."""

    def __init__(self, options):
        """Constructor.

        Args:
            options: (ClassInferenceOptions)
        """
        self.options = options

        if self.options.class_id_to_filter is not None:
            self.class_id_to_filter = {}
            for class_id, filter_exp in self.options.class_id_to_filter.items(
            ):
                self.class_id_to_filter[class_id] = create_filter(filter_exp)

    def infer_class_id(self, feature):
        """Infer the class_id for a GeoJSON feature.

        Args:
            feature: (dict) GeoJSON feature

        Rules:
            1) If class_id is in feature['properties'], use it.
            2) If class_name or label are in feature['properties'] and in class_map,
                use corresponding class_id.
            3) If class_id_to_filter is set and filter is true when applied to feature,
                use corresponding class_id.
            4) Otherwise, return the default_class_id
        """
        class_id = feature.get('properties', {}).get('class_id')
        if class_id is not None:
            return class_id

        if self.options.class_map is not None:
            class_name = feature.get('properties', {}).get('class_name')
            if class_name in self.options.class_map.get_class_names():
                return self.options.class_map.get_by_name(class_name).id

            label = feature.get('properties', {}).get('label')
            if label in self.options.class_map.get_class_names():
                return self.options.class_map.get_by_name(label).id

        if self.options.class_id_to_filter is not None:
            for class_id, filter_fn in self.class_id_to_filter.items():
                if filter_fn(feature):
                    return class_id

        return self.options.default_class_id

    def transform_geojson(self, geojson):
        """Transform GeoJSON by appending class_ids and removing features with no class.

        For each feature in geojson, the class_id is inferred and is set into
        feature['properties']. If the class_id is None (because none of the rules apply
        and the default_class_id is None), the feature is dropped.
        """
        new_features = []
        for feature in geojson['features']:
            class_id = self.infer_class_id(feature)
            if class_id is not None:
                feature = copy.deepcopy(feature)
                properties = feature.get('properties', {})
                properties['class_id'] = class_id
                feature['properties'] = properties
                new_features.append(feature)
        new_geojson = {'type': 'FeatureCollection', 'features': new_features}
        return new_geojson



from copy import deepcopy

import rastervision as rv
from rastervision.augmentor.nodata_augmentor import NodataAugmentor
from rastervision.augmentor.augmentor_config \
    import (AugmentorConfig, AugmentorConfigBuilder)
from rastervision.protos.augmentor_pb2 import AugmentorConfig as AugmentorConfigMsg


class NodataAugmentorConfig(AugmentorConfig):
    def __init__(self, aug_prob=0.5):
        super().__init__(rv.NODATA_AUGMENTOR)
        self.aug_prob = aug_prob

    def to_proto(self):
        msg = AugmentorConfigMsg(
            augmentor_type=self.augmentor_type, aug_prob=self.aug_prob)
        return msg

    def create_augmentor(self):
        return NodataAugmentor(self.aug_prob)

    def report_io(self, command_type, io_def):
        pass


class NodataAugmentorConfigBuilder(AugmentorConfigBuilder):
    def __init__(self, prev=None):
        config = {}
        if prev:
            config = {'aug_prob': prev.aug_prob}
        super().__init__(NodataAugmentorConfig, config)

    def from_proto(self, msg):
        return self.with_probablity(msg.aug_prob)

    def with_probability(self, aug_prob):
        """Sets the probability for this augmentation.

        Determines how probable this augmentation will happen
        to negative chips.

        Args:
           aug_prob: Float value between 0.0 and 1.0
        """
        b = deepcopy(self)
        b.config['aug_prob'] = aug_prob
        return b



import click

from rastervision.command import Command


class AnalyzeCommand(Command):
    def __init__(self, command_config):
        self.command_config = command_config

    def run(self, tmp_dir=None):
        if not tmp_dir:
            tmp_dir = self.get_tmp_dir()

        cc = self.command_config

        analyzers = list(map(lambda a: a.create_analyzer(), cc.analyzers))
        scenes = list(
            map(lambda s: s.create_scene(cc.task, tmp_dir), cc.scenes))

        for analyzer in analyzers:
            msg = 'Running analyzer: {}...'.format(type(analyzer).__name__)
            click.echo(click.style(msg, fg='green'))

            analyzer.process(scenes, tmp_dir)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/task.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rastervision.protos import class_item_pb2 as rastervision_dot_protos_dot_class__item__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/task.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n\x1erastervision/protos/task.proto\x12\trv.protos\x1a$rastervision/protos/class_item.proto\x1a\x1cgoogle/protobuf/struct.proto\"\x9e\x0b\n\nTaskConfig\x12\x11\n\ttask_type\x18\x01 \x02(\t\x12\x1e\n\x12predict_batch_size\x18\x02 \x01(\x05:\x02\x31\x30\x12\x1b\n\x13predict_package_uri\x18\x03 \x01(\t\x12\x13\n\x05\x64\x65\x62ug\x18\x04 \x01(\x08:\x04true\x12\x19\n\x11predict_debug_uri\x18\x05 \x01(\t\x12N\n\x17object_detection_config\x18\x06 \x01(\x0b\x32+.rv.protos.TaskConfig.ObjectDetectionConfigH\x00\x12T\n\x1a\x63hip_classification_config\x18\x07 \x01(\x0b\x32..rv.protos.TaskConfig.ChipClassificationConfigH\x00\x12X\n\x1csemantic_segmentation_config\x18\x08 \x01(\x0b\x32\x30.rv.protos.TaskConfig.SemanticSegmentationConfigH\x00\x12\x30\n\rcustom_config\x18\t \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x1a\xb2\x03\n\x15ObjectDetectionConfig\x12)\n\x0b\x63lass_items\x18\x01 \x03(\x0b\x32\x14.rv.protos.ClassItem\x12\x11\n\tchip_size\x18\x02 \x02(\x05\x12M\n\x0c\x63hip_options\x18\x03 \x02(\x0b\x32\x37.rv.protos.TaskConfig.ObjectDetectionConfig.ChipOptions\x12S\n\x0fpredict_options\x18\x04 \x02(\x0b\x32:.rv.protos.TaskConfig.ObjectDetectionConfig.PredictOptions\x1ao\n\x0b\x43hipOptions\x12\x11\n\tneg_ratio\x18\x01 \x02(\x02\x12\x17\n\nioa_thresh\x18\x02 \x01(\x02:\x03\x30.8\x12\x1b\n\rwindow_method\x18\x03 \x01(\t:\x04\x63hip\x12\x17\n\x0clabel_buffer\x18\x04 \x01(\x02:\x01\x30\x1a\x46\n\x0ePredictOptions\x12\x19\n\x0cmerge_thresh\x18\x02 \x01(\x02:\x03\x30.5\x12\x19\n\x0cscore_thresh\x18\x03 \x01(\x02:\x03\x30.5\x1aX\n\x18\x43hipClassificationConfig\x12)\n\x0b\x63lass_items\x18\x01 \x03(\x0b\x32\x14.rv.protos.ClassItem\x12\x11\n\tchip_size\x18\x02 \x02(\x05\x1a\xbf\x03\n\x1aSemanticSegmentationConfig\x12)\n\x0b\x63lass_items\x18\x01 \x03(\x0b\x32\x14.rv.protos.ClassItem\x12\x11\n\tchip_size\x18\x02 \x02(\x05\x12R\n\x0c\x63hip_options\x18\x03 \x02(\x0b\x32<.rv.protos.TaskConfig.SemanticSegmentationConfig.ChipOptions\x12\x1c\n\x11predict_chip_size\x18\x04 \x01(\x05:\x01\x30\x1a\xf0\x01\n\x0b\x43hipOptions\x12$\n\rwindow_method\x18\x01 \x01(\t:\rrandom_sample\x12\x16\n\x0etarget_classes\x18\x02 \x03(\x05\x12$\n\x16\x64\x65\x62ug_chip_probability\x18\x03 \x01(\x02:\x04\x30.25\x12(\n\x1dnegative_survival_probability\x18\x04 \x01(\x02:\x01\x31\x12\x1d\n\x0f\x63hips_per_scene\x18\x05 \x01(\x05:\x04\x31\x30\x30\x30\x12$\n\x16target_count_threshold\x18\x06 \x01(\x05:\x04\x32\x30\x34\x38\x12\x0e\n\x06stride\x18\x07 \x01(\x05\x42\r\n\x0b\x63onfig_type')
  ,
  dependencies=[rastervision_dot_protos_dot_class__item__pb2.DESCRIPTOR,google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_TASKCONFIG_OBJECTDETECTIONCONFIG_CHIPOPTIONS = _descriptor.Descriptor(
  name='ChipOptions',
  full_name='rv.protos.TaskConfig.ObjectDetectionConfig.ChipOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='neg_ratio', full_name='rv.protos.TaskConfig.ObjectDetectionConfig.ChipOptions.neg_ratio', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ioa_thresh', full_name='rv.protos.TaskConfig.ObjectDetectionConfig.ChipOptions.ioa_thresh', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.8),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='window_method', full_name='rv.protos.TaskConfig.ObjectDetectionConfig.ChipOptions.window_method', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("chip").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='label_buffer', full_name='rv.protos.TaskConfig.ObjectDetectionConfig.ChipOptions.label_buffer', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=814,
  serialized_end=925,
)

_TASKCONFIG_OBJECTDETECTIONCONFIG_PREDICTOPTIONS = _descriptor.Descriptor(
  name='PredictOptions',
  full_name='rv.protos.TaskConfig.ObjectDetectionConfig.PredictOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='merge_thresh', full_name='rv.protos.TaskConfig.ObjectDetectionConfig.PredictOptions.merge_thresh', index=0,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='score_thresh', full_name='rv.protos.TaskConfig.ObjectDetectionConfig.PredictOptions.score_thresh', index=1,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=927,
  serialized_end=997,
)

_TASKCONFIG_OBJECTDETECTIONCONFIG = _descriptor.Descriptor(
  name='ObjectDetectionConfig',
  full_name='rv.protos.TaskConfig.ObjectDetectionConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='class_items', full_name='rv.protos.TaskConfig.ObjectDetectionConfig.class_items', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chip_size', full_name='rv.protos.TaskConfig.ObjectDetectionConfig.chip_size', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chip_options', full_name='rv.protos.TaskConfig.ObjectDetectionConfig.chip_options', index=2,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='predict_options', full_name='rv.protos.TaskConfig.ObjectDetectionConfig.predict_options', index=3,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_TASKCONFIG_OBJECTDETECTIONCONFIG_CHIPOPTIONS, _TASKCONFIG_OBJECTDETECTIONCONFIG_PREDICTOPTIONS, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=563,
  serialized_end=997,
)

_TASKCONFIG_CHIPCLASSIFICATIONCONFIG = _descriptor.Descriptor(
  name='ChipClassificationConfig',
  full_name='rv.protos.TaskConfig.ChipClassificationConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='class_items', full_name='rv.protos.TaskConfig.ChipClassificationConfig.class_items', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chip_size', full_name='rv.protos.TaskConfig.ChipClassificationConfig.chip_size', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=999,
  serialized_end=1087,
)

_TASKCONFIG_SEMANTICSEGMENTATIONCONFIG_CHIPOPTIONS = _descriptor.Descriptor(
  name='ChipOptions',
  full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.ChipOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='window_method', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.ChipOptions.window_method', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("random_sample").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='target_classes', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.ChipOptions.target_classes', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='debug_chip_probability', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.ChipOptions.debug_chip_probability', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.25),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='negative_survival_probability', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.ChipOptions.negative_survival_probability', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chips_per_scene', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.ChipOptions.chips_per_scene', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='target_count_threshold', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.ChipOptions.target_count_threshold', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=2048,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stride', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.ChipOptions.stride', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1297,
  serialized_end=1537,
)

_TASKCONFIG_SEMANTICSEGMENTATIONCONFIG = _descriptor.Descriptor(
  name='SemanticSegmentationConfig',
  full_name='rv.protos.TaskConfig.SemanticSegmentationConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='class_items', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.class_items', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chip_size', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.chip_size', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chip_options', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.chip_options', index=2,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='predict_chip_size', full_name='rv.protos.TaskConfig.SemanticSegmentationConfig.predict_chip_size', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_TASKCONFIG_SEMANTICSEGMENTATIONCONFIG_CHIPOPTIONS, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1090,
  serialized_end=1537,
)

_TASKCONFIG = _descriptor.Descriptor(
  name='TaskConfig',
  full_name='rv.protos.TaskConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='task_type', full_name='rv.protos.TaskConfig.task_type', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='predict_batch_size', full_name='rv.protos.TaskConfig.predict_batch_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=10,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='predict_package_uri', full_name='rv.protos.TaskConfig.predict_package_uri', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='debug', full_name='rv.protos.TaskConfig.debug', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='predict_debug_uri', full_name='rv.protos.TaskConfig.predict_debug_uri', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='object_detection_config', full_name='rv.protos.TaskConfig.object_detection_config', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chip_classification_config', full_name='rv.protos.TaskConfig.chip_classification_config', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='semantic_segmentation_config', full_name='rv.protos.TaskConfig.semantic_segmentation_config', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='custom_config', full_name='rv.protos.TaskConfig.custom_config', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_TASKCONFIG_OBJECTDETECTIONCONFIG, _TASKCONFIG_CHIPCLASSIFICATIONCONFIG, _TASKCONFIG_SEMANTICSEGMENTATIONCONFIG, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='config_type', full_name='rv.protos.TaskConfig.config_type',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=114,
  serialized_end=1552,
)

_TASKCONFIG_OBJECTDETECTIONCONFIG_CHIPOPTIONS.containing_type = _TASKCONFIG_OBJECTDETECTIONCONFIG
_TASKCONFIG_OBJECTDETECTIONCONFIG_PREDICTOPTIONS.containing_type = _TASKCONFIG_OBJECTDETECTIONCONFIG
_TASKCONFIG_OBJECTDETECTIONCONFIG.fields_by_name['class_items'].message_type = rastervision_dot_protos_dot_class__item__pb2._CLASSITEM
_TASKCONFIG_OBJECTDETECTIONCONFIG.fields_by_name['chip_options'].message_type = _TASKCONFIG_OBJECTDETECTIONCONFIG_CHIPOPTIONS
_TASKCONFIG_OBJECTDETECTIONCONFIG.fields_by_name['predict_options'].message_type = _TASKCONFIG_OBJECTDETECTIONCONFIG_PREDICTOPTIONS
_TASKCONFIG_OBJECTDETECTIONCONFIG.containing_type = _TASKCONFIG
_TASKCONFIG_CHIPCLASSIFICATIONCONFIG.fields_by_name['class_items'].message_type = rastervision_dot_protos_dot_class__item__pb2._CLASSITEM
_TASKCONFIG_CHIPCLASSIFICATIONCONFIG.containing_type = _TASKCONFIG
_TASKCONFIG_SEMANTICSEGMENTATIONCONFIG_CHIPOPTIONS.containing_type = _TASKCONFIG_SEMANTICSEGMENTATIONCONFIG
_TASKCONFIG_SEMANTICSEGMENTATIONCONFIG.fields_by_name['class_items'].message_type = rastervision_dot_protos_dot_class__item__pb2._CLASSITEM
_TASKCONFIG_SEMANTICSEGMENTATIONCONFIG.fields_by_name['chip_options'].message_type = _TASKCONFIG_SEMANTICSEGMENTATIONCONFIG_CHIPOPTIONS
_TASKCONFIG_SEMANTICSEGMENTATIONCONFIG.containing_type = _TASKCONFIG
_TASKCONFIG.fields_by_name['object_detection_config'].message_type = _TASKCONFIG_OBJECTDETECTIONCONFIG
_TASKCONFIG.fields_by_name['chip_classification_config'].message_type = _TASKCONFIG_CHIPCLASSIFICATIONCONFIG
_TASKCONFIG.fields_by_name['semantic_segmentation_config'].message_type = _TASKCONFIG_SEMANTICSEGMENTATIONCONFIG
_TASKCONFIG.fields_by_name['custom_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_TASKCONFIG.oneofs_by_name['config_type'].fields.append(
  _TASKCONFIG.fields_by_name['object_detection_config'])
_TASKCONFIG.fields_by_name['object_detection_config'].containing_oneof = _TASKCONFIG.oneofs_by_name['config_type']
_TASKCONFIG.oneofs_by_name['config_type'].fields.append(
  _TASKCONFIG.fields_by_name['chip_classification_config'])
_TASKCONFIG.fields_by_name['chip_classification_config'].containing_oneof = _TASKCONFIG.oneofs_by_name['config_type']
_TASKCONFIG.oneofs_by_name['config_type'].fields.append(
  _TASKCONFIG.fields_by_name['semantic_segmentation_config'])
_TASKCONFIG.fields_by_name['semantic_segmentation_config'].containing_oneof = _TASKCONFIG.oneofs_by_name['config_type']
_TASKCONFIG.oneofs_by_name['config_type'].fields.append(
  _TASKCONFIG.fields_by_name['custom_config'])
_TASKCONFIG.fields_by_name['custom_config'].containing_oneof = _TASKCONFIG.oneofs_by_name['config_type']
DESCRIPTOR.message_types_by_name['TaskConfig'] = _TASKCONFIG

TaskConfig = _reflection.GeneratedProtocolMessageType('TaskConfig', (_message.Message,), dict(

  ObjectDetectionConfig = _reflection.GeneratedProtocolMessageType('ObjectDetectionConfig', (_message.Message,), dict(

    ChipOptions = _reflection.GeneratedProtocolMessageType('ChipOptions', (_message.Message,), dict(
      DESCRIPTOR = _TASKCONFIG_OBJECTDETECTIONCONFIG_CHIPOPTIONS,
      __module__ = 'rastervision.protos.task_pb2'
      # @@protoc_insertion_point(class_scope:rv.protos.TaskConfig.ObjectDetectionConfig.ChipOptions)
      ))
    ,

    PredictOptions = _reflection.GeneratedProtocolMessageType('PredictOptions', (_message.Message,), dict(
      DESCRIPTOR = _TASKCONFIG_OBJECTDETECTIONCONFIG_PREDICTOPTIONS,
      __module__ = 'rastervision.protos.task_pb2'
      # @@protoc_insertion_point(class_scope:rv.protos.TaskConfig.ObjectDetectionConfig.PredictOptions)
      ))
    ,
    DESCRIPTOR = _TASKCONFIG_OBJECTDETECTIONCONFIG,
    __module__ = 'rastervision.protos.task_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.TaskConfig.ObjectDetectionConfig)
    ))
  ,

  ChipClassificationConfig = _reflection.GeneratedProtocolMessageType('ChipClassificationConfig', (_message.Message,), dict(
    DESCRIPTOR = _TASKCONFIG_CHIPCLASSIFICATIONCONFIG,
    __module__ = 'rastervision.protos.task_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.TaskConfig.ChipClassificationConfig)
    ))
  ,

  SemanticSegmentationConfig = _reflection.GeneratedProtocolMessageType('SemanticSegmentationConfig', (_message.Message,), dict(

    ChipOptions = _reflection.GeneratedProtocolMessageType('ChipOptions', (_message.Message,), dict(
      DESCRIPTOR = _TASKCONFIG_SEMANTICSEGMENTATIONCONFIG_CHIPOPTIONS,
      __module__ = 'rastervision.protos.task_pb2'
      # @@protoc_insertion_point(class_scope:rv.protos.TaskConfig.SemanticSegmentationConfig.ChipOptions)
      ))
    ,
    DESCRIPTOR = _TASKCONFIG_SEMANTICSEGMENTATIONCONFIG,
    __module__ = 'rastervision.protos.task_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.TaskConfig.SemanticSegmentationConfig)
    ))
  ,
  DESCRIPTOR = _TASKCONFIG,
  __module__ = 'rastervision.protos.task_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.TaskConfig)
  ))
_sym_db.RegisterMessage(TaskConfig)
_sym_db.RegisterMessage(TaskConfig.ObjectDetectionConfig)
_sym_db.RegisterMessage(TaskConfig.ObjectDetectionConfig.ChipOptions)
_sym_db.RegisterMessage(TaskConfig.ObjectDetectionConfig.PredictOptions)
_sym_db.RegisterMessage(TaskConfig.ChipClassificationConfig)
_sym_db.RegisterMessage(TaskConfig.SemanticSegmentationConfig)
_sym_db.RegisterMessage(TaskConfig.SemanticSegmentationConfig.ChipOptions)


# @@protoc_insertion_point(module_scope)



import numpy as np
import shapely
import shapely.strtree
import shapely.geometry

from rastervision.data import ObjectDetectionLabels
from rastervision.evaluation import ClassEvaluationItem
from rastervision.evaluation import ClassificationEvaluation


def compute_metrics(gt_labels: ObjectDetectionLabels,
                    pred_labels: ObjectDetectionLabels,
                    num_classes: int,
                    iou_thresh=0.5):
    gt_geoms = [b.to_shapely() for b in gt_labels.get_boxes()]
    gt_classes = gt_labels.get_class_ids() - 1
    pred_geoms = [b.to_shapely() for b in pred_labels.get_boxes()]
    pred_classes = pred_labels.get_class_ids() - 1

    for pred, class_id in zip(pred_geoms, pred_classes):
        pred.class_id = class_id
    pred_tree = shapely.strtree.STRtree(pred_geoms)

    def iou(a, b):
        return a.intersection(b).area / a.union(b).area

    def is_matched(geom):
        return hasattr(geom, 'iou_matched')

    tp = np.zeros((num_classes, ))
    fp = np.zeros((num_classes, ))
    fn = np.zeros((num_classes, ))

    for gt, gt_class in zip(gt_geoms, gt_classes):
        matches = list(
            filter(lambda g: (not is_matched(g)) and g.class_id == gt_class,
                   pred_tree.query(gt)))
        scores = [iou(m, gt) for m in matches]
        if len(scores) > 0:
            max_ind = np.argmax(scores)
            if scores[max_ind] > iou_thresh:
                matches[max_ind].iou_matched = True
                tp[gt_class] += 1
            else:
                fn[gt_class] += 1
        else:
            fn[gt_class] += 1

    for class_id in range(num_classes):
        pred_not_matched = np.array([not is_matched(g) for g in pred_geoms])
        fp[class_id] = np.sum(pred_not_matched[pred_classes == class_id])

    return tp, fp, fn


class ObjectDetectionEvaluation(ClassificationEvaluation):
    def __init__(self, class_map):
        super().__init__()
        self.class_map = class_map

    def compute(self, ground_truth_labels, prediction_labels):
        self.class_to_eval_item = ObjectDetectionEvaluation.compute_eval_items(
            ground_truth_labels, prediction_labels, self.class_map)
        self.compute_avg()

    @staticmethod
    def compute_eval_items(gt_labels, pred_labels, class_map):
        iou_thresh = 0.5
        num_classes = len(class_map)
        tps, fps, fns = compute_metrics(gt_labels, pred_labels, num_classes,
                                        iou_thresh)
        class_to_eval_item = {}

        for class_ind, (tp, fp, fn) in enumerate(zip(tps, fps, fns)):
            class_id = class_ind + 1
            gt_count = tp + fn
            pred_count = tp + fp
            class_name = class_map.get_by_id(class_id).name

            if gt_count == 0:
                eval_item = ClassEvaluationItem(
                    class_id=class_id, class_name=class_name)
            elif pred_count == 0:
                eval_item = ClassEvaluationItem(
                    precision=None,
                    recall=0,
                    gt_count=gt_count,
                    class_id=class_id,
                    class_name=class_name)
            else:
                prec = tp / (tp + fp)
                recall = tp / (tp + fn)
                f1 = 0.
                if prec + recall != 0.0:
                    f1 = 2 * (prec * recall) / (prec + recall)
                count_err = pred_count - gt_count
                norm_count_err = None
                if gt_count > 0:
                    norm_count_err = count_err / gt_count

                eval_item = ClassEvaluationItem(
                    precision=prec,
                    recall=recall,
                    f1=f1,
                    count_error=norm_count_err,
                    gt_count=gt_count,
                    class_id=class_id,
                    class_name=class_name)

            class_to_eval_item[class_id] = eval_item

        return class_to_eval_item
